# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import List

from msprobe.core.compare.indicator_analysis.utils import CompareMode, ResultLevel, str2float
from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.log import logger
from msprobe.core.common.exceptions import MsprobeException


class ApiData:
    header_index_mapping_cache = {}

    def __init__(self, mode, data_lists: List[List]):
        self.input_data = []
        self.output_data = []
        self.mode = mode
        self.data_lists = data_lists
        self.header = self._get_header()
        # 表头与索引映射，可以基于表头拿到索引，从input_data和output_data中获取对应数据
        self.header_index_mapping = self.get_header_index_mapping()
        self._init_data()

    def get_header_index_mapping(self):
        if self.mode in ApiData.header_index_mapping_cache:
            return ApiData.header_index_mapping_cache[self.mode]

        mapping = {item: index for index, item in enumerate(self.header)}
        ApiData.header_index_mapping_cache[self.mode] = mapping
        return mapping

    def get_data_by_header(self, header: str, data_list: List):
        """
        基于表头从data list获取数据
        """
        index = self.header_index_mapping.get(header)
        try:
            data = data_list[index]
        except Exception as e:
            logger.error(f'Unable to get data from the data list based on the header: {e}')
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR) from e
        return data

    def set_data_by_header(self, header: str, data_list: List, new_data):
        index = self.header_index_mapping.get(header)
        try:
            data_list[index] = new_data
        except Exception as e:
            logger.error(f'Unable to set data from the data list based on the header: {e}')
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR) from e

    def set_result(self, data_list: List, status: ResultLevel = ResultLevel.PASS):
        index = self.header_index_mapping.get(CompareConst.RESULT)
        try:
            current_status = data_list[index]
            if not isinstance(current_status, ResultLevel) or status > current_status:
                data_list[index] = status
        except Exception as e:
            logger.error(f'Unable to set status from the data list based on the header: {e}')
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR) from e

    def set_err_msg(self, data_list: List, msg: str = '', init_msg=False):
        index = self.header_index_mapping.get(CompareConst.ERROR_MESSAGE)
        try:
            if init_msg:
                data_list[index] = []
            else:
                current_msg = data_list[index]
                if not isinstance(current_msg, list):
                    current_msg = [current_msg] if current_msg else []
                    data_list[index] = current_msg
                if msg:
                    current_msg.append(msg)
        except Exception as e:
            logger.error(f'Unable to set err msg from the data list based on the header: {e}')
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR) from e

    def get_min_or_max_value(self, header, is_input=True, is_min=True):
        """
        获取多个输入或输出参数中的最小或最大指标
        """
        default_value = 1.0 if is_min else 0.0
        data_lists = self.input_data if is_input else self.output_data
        for data_list in data_lists:
            value = self.get_data_by_header(header, data_list)

            if value is None or value in [CompareConst.NAN, CompareConst.N_A]:
                continue

            if isinstance(value, str) and value.endswith('%'):
                value = str2float(value)

            if isinstance(value, str):
                continue

            default_value = min(default_value, value) if is_min else max(default_value, value)
        return default_value

    def _get_header(self):
        if self.mode == CompareMode.STATISTICS.value:
            return CompareConst.SUMMARY_COMPARE_RESULT_HEADER
        elif self.mode == CompareMode.TENSOR.value:
            return CompareConst.COMPARE_RESULT_HEADER
        elif self.mode == CompareMode.MD5.value:
            return CompareConst.MD5_COMPARE_RESULT_HEADER
        else:
            logger.error(f'The parameter "mode" error, '
                         f'expected {CompareMode.STATISTICS.value}/{CompareMode.TENSOR.value}/{CompareMode.MD5.value}, '
                         f'actually {self.mode}.')
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    def _init_data(self):
        for data_list in self.data_lists:
            if not data_list:
                continue

            if len(data_list) < len(self.header):
                data_list = data_list + [''] * (len(self.header) - len(data_list))

            # 初始化result为pass，error message为list
            self.set_result(data_list)
            self.set_err_msg(data_list, init_msg=True)

            if f'{Const.SEP}{Const.OUTPUT}{Const.SEP}' in data_list[0]:
                self.output_data.append(data_list)
            else:
                self.input_data.append(data_list)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch


def npu_apply_adam_w(beta1_power, beta2_power, lr, weight_decay,
                     beta1, beta2, eps, grad, max_grad_norm, amsgrad, maximize, out):
    var, m, v = out
    if amsgrad:
        max_grad_norm = (torch.rand(var.shape) * 10.0 - 5.0).to(var.dtype)
    beta1_power_out = beta1_power * beta1
    beta2_power_out = beta2_power * beta2
    var_t = var * (1 + (-lr * weight_decay))
    gt = -grad if maximize else grad
    m_out = m * beta1 - (beta1 + (-1)) * gt
    v_out = v * beta2 - (beta2 + (-1)) * gt * gt

    if amsgrad:
        max_grad_norm_out = torch.max(max_grad_norm, v_out)
        if (1 - beta2_power_out) == 0:
            beta2_power_out -= eps
        denom = torch.sqrt(torch.div(max_grad_norm_out, (1 - beta2_power_out))) + eps
    else:
        vraintain = torch.div(v_out, (1 - beta2_power_out))
        denom = torch.sqrt(vraintain) + eps

    if (1 - beta1_power_out) == 0:
        beta1_power_out -= eps
    var_out = var_t + torch.div(-lr * m_out, (1 - beta1_power_out)).div(denom)
    return var_out, m_out, v_out



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.file_utils import check_path_pattern_valid
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.const import FileCheckConst


class PackInput:

    def __init__(self, output_zip_path, model, shell_path):
        self.output_zip_path = output_zip_path
        self.shell_path = shell_path
        self.model = model[0] if isinstance(model, list) and len(model) > 0 else model
        self.check_input_params()

    def check_input_params(self):
        if self.model and not FmkAdp.is_nn_module(self.model):
            raise Exception(f"model is not torch.nn.Module/mindspore.nn.Cell or module list.")
        if not isinstance(self.output_zip_path, str) or not self.output_zip_path.endswith(FileCheckConst.ZIP_SUFFIX):
            raise Exception(f"output zip path must be a string and ends with '.zip'")
        check_path_pattern_valid(self.output_zip_path)


class BaseChecker:
    input_needed = None
    target_name_in_zip = None
    multi_rank = False

    @staticmethod
    def pack(pack_input):
        pass

    @staticmethod
    def compare(bench_dir, cmp_dir, output_path):
        pass

    @staticmethod
    def apply_patches(fmk):
        pass

    @classmethod
    def compare_ex(cls, bench_dir, cmp_dir, output_path):
        bench_filepath = os.path.join(bench_dir, cls.target_name_in_zip)
        cmp_filepath = os.path.join(cmp_dir, cls.target_name_in_zip)
        if not os.path.exists(bench_filepath) or not os.path.exists(cmp_filepath):
            return None, None, None
        return cls.compare(bench_dir, cmp_dir, output_path)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import threading
from collections import OrderedDict

import mindspore as ms
from mindspore import Tensor
from mindspore.common.api import _no_grad, _pynative_executor
from mindspore.ops.operations import _inner_ops as inner

from msprobe.core.common.const import Const
from msprobe.core.common.log import logger
from msprobe.core.common.utils import replace_last_occurrence, ThreadSafe
from msprobe.core.dump.data_dump.data_processor.base import ModuleBackwardInputs
from msprobe.core.dump.hook_manager import BaseHookManager, HookSet
from msprobe.mindspore.common.const import Const as MsConst
from msprobe.mindspore.common.utils import (
    has_kwargs_in_forward_hook,
    is_mindtorch,
    is_backward_hook_output_a_view
)
from msprobe.mindspore.dump.dump_processor.hook_cell.hook_cell import HOOKCell

ms_version = ms.__version__


class MindsporeHookManager(BaseHookManager):
    cell_bw_hook_kernels = {}
    cell_backward_pre_hook = []
    cell_backward_hook = []

    @staticmethod
    def reset_status():
        BaseHookManager.reset_status()
        MindsporeHookManager.cell_bw_hook_kernels.clear()
        MindsporeHookManager.cell_backward_pre_hook.clear()
        MindsporeHookManager.cell_backward_hook.clear()

    @staticmethod
    def _no_grad_context():
        return _no_grad()

    @staticmethod
    def _add_count(name):
        HOOKCell.add_cell_count(name)

    @staticmethod
    def _get_count(name):
        return HOOKCell.get_cell_count(name)

    @staticmethod
    def _process_kwargs_and_output(module, tid, hook_type, kwargs_or_output, output_or_kwargs):
        if not has_kwargs_in_forward_hook() or hook_type == Const.API:
            kwargs = module.msprobe_input_kwargs.get(tid, {}) if hasattr(module, 'msprobe_input_kwargs') else {}
            output = kwargs_or_output
        else:
            kwargs = kwargs_or_output
            output = output_or_kwargs
        return kwargs, output

    def build_hook(self, hook_type, name):
        if hook_type == Const.API:
            hook_set = HookSet(
                forward_pre_hook=self._build_forward_pre_hook(hook_type, name)
            )
        else:
            full_backward_name = replace_last_occurrence(name, Const.FORWARD, Const.BACKWARD)
            hook_set = HookSet(
                forward_hook=self._build_forward_hook(hook_type, name),
                backward_pre_hook=self._build_backward_pre_hook(hook_type, full_backward_name),
                backward_hook=self._build_backward_hook(hook_type, full_backward_name)
            )
        return hook_set

    def _register_forward_hook(self, module, api_name):
        if not hasattr(module, 'msprobe_forward_hook'):
            forward_hook = self._build_forward_hook(Const.API, api_name)
            if ms_version < "2.6.0" and not is_mindtorch():
                getattr(module, "_forward_hook", {})[id(module)] = forward_hook
            else:
                module.register_forward_hook(forward_hook)
            setattr(module, 'msprobe_forward_hook', True)

    def _register_backward_hook(self, module, full_backward_name, args):
        if not _pynative_executor.requires_grad():
            return args

        enable_hooked = sum(
            [isinstance(ele, Tensor) and ele.dtype not in MsConst.NonDifferentiableType for ele in args]
        )

        if enable_hooked:
            backward_hook_dict = OrderedDict()
            backward_hook_dict[full_backward_name] = self._build_backward_hook(Const.API, full_backward_name)
            MindsporeHookManager.cell_backward_hook.append(backward_hook_dict)
            bw_hook = inner.CellBackwardHook(full_backward_name, module, MindsporeHookManager.cell_backward_hook[-1])
            bw_hook.register_backward_hook()
            MindsporeHookManager.cell_bw_hook_kernels[full_backward_name] = bw_hook
            args = bw_hook(args) if is_backward_hook_output_a_view() else bw_hook(*args)
        return args

    def _register_backward_pre_hook(self, module, full_backward_name, args, kwargs, output):
        if not _pynative_executor.requires_grad():
            return output

        bw_hook = MindsporeHookManager.cell_bw_hook_kernels.get(full_backward_name)
        if bw_hook:
            if not isinstance(output, (Tensor, tuple)):
                logger.debug("For backward hooks to be called, "
                             "cell output should be a Tensor or a tuple of Tensors "
                             f"but received {type(output)}")
            if is_backward_hook_output_a_view():
                new_outputs = bw_hook(output)
            else:
                if isinstance(output, tuple):
                    new_outputs = bw_hook(*output)
                else:
                    new_outputs = bw_hook(output)
                if isinstance(output, tuple) and len(output) == 1:
                    new_outputs = (new_outputs,)
            output = new_outputs

        def get_backward_pre_hook(backward_pre_hook, backward_post_hook):
            @ThreadSafe.synchronized
            def backward_pre_hook_fn(cell, grad_output):
                backward_pre_hook(cell, grad_output)
                if backward_post_hook:
                    backward_post_hook(cell, (), grad_output)

            return backward_pre_hook_fn

        backward_pre_hook = self._build_backward_pre_hook(Const.API, full_backward_name)
        backward_post_hook = None if bw_hook else self._build_backward_hook(Const.API, full_backward_name)

        backward_pre_hook_dict = OrderedDict()
        backward_pre_hook_dict[full_backward_name] = get_backward_pre_hook(
            backward_pre_hook,
            backward_post_hook
        )
        MindsporeHookManager.cell_backward_pre_hook.append(backward_pre_hook_dict)
        bw_pre_hook = inner.CellBackwardHook(
            full_backward_name,
            module,
            MindsporeHookManager.cell_backward_pre_hook[-1]
        )
        bw_pre_hook.register_backward_pre_hook()

        if is_backward_hook_output_a_view():
            result = bw_pre_hook(output)
        else:
            if isinstance(output, tuple):
                result = bw_pre_hook(*output)
            else:
                result = bw_pre_hook(output)
            if isinstance(output, tuple):
                if len(output) == 1:
                    result = (result,)
                if len(result) != len(output):
                    raise TypeError(
                        f"The backward pre hook return value size is {len(result)} "
                        f"not equal to output size {len(output)}"
                    )
        return result

    def _need_exchange(self, module):
        if not hasattr(module, 'has_pre_hook_called') or not module.has_pre_hook_called:
            return False
        else:
            return True

    def _get_params_dict(self, module):
        params_dict = {}
        if self.config.task != Const.STRUCTURE:
            params_dict = {
                key.split(Const.SEP)[-1]: value
                for key, value in module.parameters_dict(recurse=False).items()
            }
        return params_dict

    def _build_backward_pre_hook(self, hook_type, full_name):
        def backward_pre_hook(module, grad_input):
            if self.config.level != Const.LEVEL_L2:
                return
            tid = threading.get_ident()
            if not self._should_execute_hook(hook_type, tid):
                return

            with ThreadSafe():
                original_state = self.ensure_gc_enabled()
                BaseHookManager.inner_switch[tid] = True
                module_input = ModuleBackwardInputs(grad_input=grad_input)
                self.data_collector.update_api_or_module_name(full_name)
                self.data_collector.backward_input_data_collect(full_name, module, self._pid, module_input)
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)

        return backward_pre_hook

    def _register_param_hook(self, name, module, params_dict):
        ori_name = name.rsplit(Const.SEP, 2)[0]
        # data_mode为forward时，不注册参数hook
        if not (Const.FORWARD in self.config.data_mode and Const.BACKWARD not in self.config.data_mode):
            for param_name, param in params_dict.items():
                if param.requires_grad:
                    name = ori_name + Const.SEP + param_name
                    old_handle = BaseHookManager.hook_handle_dict.get(name)
                    if old_handle and hasattr(old_handle, "remove"):
                        old_handle.remove()
                    handle = param.register_hook(self._build_grad_hook(ori_name, param_name))
                    BaseHookManager.hook_handle_dict[name] = handle

    def _build_grad_hook(self, ori_name, param_name):
        def hook_fn(grad):
            tid = threading.get_ident()
            if not self._should_execute_hook(Const.MODULE, tid):
                return
            with ThreadSafe():
                original_state = self.ensure_gc_enabled()
                BaseHookManager.inner_switch[tid] = True
                index = self._get_grad_hook_call_index(ori_name, param_name)
                self.data_collector.params_data_collect(ori_name, param_name, self._pid, grad, str(index))
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)
            return

        return hook_fn


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
import multiprocessing
from functools import partial

import pandas as pd
from tqdm import tqdm

from msprobe.core.common.file_utils import load_yaml, logger, FileChecker, save_excel, read_xlsx, create_directory, \
    remove_path
from msprobe.core.common.const import FileCheckConst, Const, CompareConst
from msprobe.core.common.utils import CompareException, add_time_with_xlsx
from msprobe.core.compare.utils import table_value_is_valid
from msprobe.core.compare.merge_result.utils import replace_compare_index_dict, check_config


def check_compare_result_name(file_name):
    """
    check whether the compare result name is as expected
    """
    single_rank_pattern = r"^compare_result_(rank|rank-rank)_\d{14}\.xlsx$"
    multi_ranks_pattern = r"^compare_result_rank(\d+)(?:-rank\1)?_\d{14}\.xlsx$"
    if re.match(multi_ranks_pattern, file_name):
        return True
    if re.match(single_rank_pattern, file_name):
        logger.warning("Single rank compare result do not need to be merged.")
        return False
    logger.error(f"Wrong compare result name: {file_name}, please check!")
    raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)


def reorder_path(compare_result_path_list):
    """
    reorder compare results by rank num
    """
    rank_pattern = r"compare_result_rank(\d+)"
    reorder_path_list = sorted(
        compare_result_path_list,
        key=lambda path: int(re.search(rank_pattern, os.path.basename(path)).group(1))
    )
    return reorder_path_list


def get_result_path(input_dir):
    """
    get rank ordered compare result file path list
    """
    compare_result_path_list = [os.path.join(input_dir, f)
                                for f in os.listdir(input_dir) if f.endswith(FileCheckConst.XLSX_SUFFIX)]
    filt_compare_result_path_list = []
    for file_path in compare_result_path_list:
        FileChecker(file_path, FileCheckConst.FILE, FileCheckConst.READ_ABLE).common_check()
        file_name = os.path.basename(file_path)
        if check_compare_result_name(file_name):
            compare_result_path_checker = FileChecker(file_path, FileCheckConst.FILE, FileCheckConst.READ_ABLE)
            compare_result_path = compare_result_path_checker.common_check()
            filt_compare_result_path_list.append(compare_result_path)

    filt_compare_result_path_list = reorder_path(filt_compare_result_path_list)       # 多卡比对结果按rank序号重新排序

    if len(filt_compare_result_path_list) < 2:
        logger.warning("Number of compare result is no more than 1, no need to merge.")     # 单卡结果无需合并，直接退出
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)
    return filt_compare_result_path_list


def get_dump_mode(result_df, rank_num):

    """
    get dump mode from header of first compare result table
    """
    header = result_df.columns.tolist()
    if header in [CompareConst.COMPARE_RESULT_HEADER + [CompareConst.DATA_NAME],
                  CompareConst.COMPARE_RESULT_HEADER_STACK + [CompareConst.DATA_NAME]]:
        return Const.ALL
    elif header in [CompareConst.SUMMARY_COMPARE_RESULT_HEADER, CompareConst.SUMMARY_COMPARE_RESULT_HEADER_STACK]:
        return Const.SUMMARY
    elif header in [CompareConst.MD5_COMPARE_RESULT_HEADER, CompareConst.MD5_COMPARE_RESULT_HEADER_STACK]:
        return Const.MD5
    else:
        logger.warning(f"A valid dump task can not be identified from rank{rank_num} compare result, please check! "
                       f"The compare result will not be shown in merged result.")
        return ""


def check_index_dump_mode_consistent(dump_mode, rank_num):
    """
    check compare index to merge is consistent with dump mode
    if compare_index_list is None, return all compare_indexes of dump mode
    """
    if dump_mode == Const.MD5:
        logger.warning(f"Rank{rank_num} compare result is 'md5' dump task and does not support merging result, please "
                       f"check! The compare result will not be shown in merged result.")
        return []

    dump_mode_compare_index_map = {
        Const.ALL: CompareConst.ALL_COMPARE_INDEX + [CompareConst.REQ_GRAD_CONSIST],
        Const.SUMMARY: CompareConst.SUMMARY_COMPARE_INDEX + [CompareConst.REQ_GRAD_CONSIST]
    }
    valid_compare_index = dump_mode_compare_index_map.get(dump_mode)

    share_list = list(share_compare_index_list)

    # 如果传入的compare_index_list为空，则比对指标为dump_mode对应的全部比对指标
    if not share_list:
        share_compare_index_list.extend(valid_compare_index)
        return list(share_compare_index_list)
    if set(share_list).issubset(valid_compare_index):
        return share_list
    else:
        invalid_compare_index = set(valid_compare_index) - set(share_list)
        logger.warning(f"Compare indexes in rank{rank_num} compare result are not consistent with "
                       f"those in other compare results, please check!")
        logger.warning(f"The compare result will not be shown in merged result.")
        logger.warning(f"The invalid compare indexes: {invalid_compare_index}")
        return []


def extract_api_full_name(api_list, result_df, rank_num):
    """
    find api full name from compare result according to api list
    """
    api_full_name_list = []
    for api in api_list:
        api_pat = api + Const.SEP
        escaped_api_pat = api_pat.replace('.', r'\.')
        single_api_full_name_list = result_df.loc[
            result_df[CompareConst.NPU_NAME].str.contains(escaped_api_pat, na=False), CompareConst.NPU_NAME].tolist()
        if len(single_api_full_name_list) == 0:
            logger.warning(f"{api} not found in rank{rank_num} compare result.")
            continue
        api_full_name_list.extend(single_api_full_name_list)
    return api_full_name_list


def search_api_index_result(api_list, compare_index_list, result_df, rank_num, compare_index_dict):
    """
    parsing single rank compare result into the intermediate target dict
    {
        compare_index1: {
            api_full_name1:{
                rank1: value,
            },
            api_full_name2,
            ...
        },
        compare_index2: {},
        ...
    }
    """
    api_full_name_list = extract_api_full_name(api_list, result_df, rank_num)
    for compare_index in compare_index_list:
        api_index_dict = {}
        for api_full_name in api_full_name_list:
            table_value_check(api_full_name)
            row_num = result_df.index[result_df[CompareConst.NPU_NAME] == api_full_name].tolist()[0]
            index_value = result_df.loc[row_num, compare_index]
            table_value_check(index_value)
            api_index_dict.setdefault(api_full_name, {})[rank_num] = index_value  # update api_index_dict
        compare_index_dict[compare_index] = api_index_dict

    compare_index_dict = replace_compare_index_dict(compare_index_dict, compare_index_list, rank_num)
    return compare_index_dict


def table_value_check(value):
    if not table_value_is_valid(value):
        raise RuntimeError(
            f"Malicious value [{value}] is not allowed to be written into the merged xlsx.")


def result_process(compare_result_path_list, api_list):
    """
    process compare results into target intermediate dict list
    """
    compare_index_dict_list = []
    rank_num_list = []
    compare_index_list = []

    for compare_result_path in compare_result_path_list:
        compare_index_dict = {}
        result_df = read_xlsx(compare_result_path)

        rank_pattern = r"compare_result_rank(\d+)"
        rank_num = int(re.search(rank_pattern, os.path.basename(compare_result_path)).group(1))
        logger.info(f"Parsing rank{rank_num} compare result...")
        if not result_df.empty:
            dump_mode = get_dump_mode(result_df, rank_num)
            if dump_mode == "":
                return [], [], []
            # 因为compare_index是指定的，固定不变，所以一旦compare_index是确定的，dump_mode也是确定的，
            # 所以只要校验compare_index和dump_mode一致性就能保证所有rank的结果都是dump_mode一致的
            compare_index_list = check_index_dump_mode_consistent(dump_mode, rank_num)
            if len(compare_index_list) == 0:
                return [], [], []
            compare_index_list.extend([CompareConst.NPU_MAX, CompareConst.BENCH_MAX])
            compare_index_dict = search_api_index_result(api_list, compare_index_list,
                                                         result_df, rank_num, compare_index_dict)
            compare_index_dict_list.append(compare_index_dict)
            rank_num_list.append(rank_num)
            compare_index_list.pop()
            compare_index_list.pop()
        else:
            logger.warning(f"Rank{rank_num} compare result is empty and will not shown in merged result.")

    return compare_index_dict_list, rank_num_list, compare_index_list


def handle_multi_process(func, func_args, lock):
    compare_result_path_list, api_list = func_args

    result_num = len(compare_result_path_list)
    process_num = int((multiprocessing.cpu_count() + 1) / 2)
    if result_num <= process_num:
        process_num = result_num
        chunks = [[compare_result_path] for compare_result_path in compare_result_path_list]
    else:
        chunk_size = result_num // process_num
        chunks = [compare_result_path_list[i:i + chunk_size] for i in range(0, result_num, chunk_size)]

    pool = multiprocessing.Pool(process_num)

    def err_call(args):
        logger.error('Multiprocess merge result failed! Reason: {}'.format(args))
        try:
            pool.close()
        except OSError:
            logger.error("Pool terminate failed")

    progress_bar = tqdm(total=result_num, desc="Compare Result Parsing Process", unit="num", ncols=100)

    def update_progress(size, progress_lock, extra_param=None):
        with progress_lock:
            progress_bar.update(size)

    results = []
    for chunk in chunks:
        chunk_size = len(chunk)
        result = pool.apply_async(func,     # pool.apply_async立即返回ApplyResult对象，因此results中结果是顺序的
                                  args=(chunk, api_list),
                                  error_callback=err_call,
                                  callback=partial(update_progress, chunk_size, lock)
                                  )
        results.append(result)

    all_compare_index_dict_list = []
    all_rank_num_list = []
    all_compare_index_list_list = []
    for result in results:
        compare_index_dict, rank_num_list, compare_index_list = result.get()
        all_compare_index_dict_list.append(compare_index_dict)
        all_rank_num_list.append(rank_num_list)
        all_compare_index_list_list.append(compare_index_list)

    pool.close()
    pool.join()

    if not any(all_compare_index_dict_list):
        logger.warning("Nothing to merge.")
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)

    return all_compare_index_dict_list, all_rank_num_list, all_compare_index_list_list


def generate_result_df(api_index_dict, header):
    """
    Generates a DataFrame from the given api_index_dict and header.
    api_index_dict:
    {
        api_full_name1:{
            rank1: value,
            },
        api_full_name2:{
            rank1: value
            },
        ...
    }
    """
    result = []
    for api_full_name, rank_value_dict in api_index_dict.items():
        result_item = [api_full_name]
        result_item.extend(rank_value_dict.values())
        result.append(result_item)
    return pd.DataFrame(result, columns=header, dtype="object")


def generate_merge_result(all_compare_index_dict_list, all_rank_num_list, all_compare_index_list_list, output_dir):
    """
    generate merge result from the intermediate dict.
    one compare index, one sheet
    """
    file_name = add_time_with_xlsx("multi_ranks_compare_merge")
    output_path = os.path.join(output_dir, file_name)

    compare_index_list = None
    for item in all_compare_index_list_list:
        if item:
            compare_index_list = item
            break
    if not compare_index_list:
        logger.error("No compare index recognized, please check!")
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)

    all_result_df_list = []
    for compare_index_dict_list, rank_num_list in zip(all_compare_index_dict_list, all_rank_num_list):
        for compare_index_dict, rank_num in zip(compare_index_dict_list, rank_num_list):
            header = [CompareConst.NPU_NAME, "rank" + str(rank_num)]
            result_df_list = []
            for _, api_index_dict in compare_index_dict.items():
                result_df = generate_result_df(api_index_dict, header)
                result_df_list.append(result_df)
            all_result_df_list.append(result_df_list)

    merge_df_list = df_merge(all_result_df_list)
    final_result_df_list = []
    for i, df in enumerate(merge_df_list):
        # merge_df_list中df与compare_index_list中compare_index一一对应
        final_result_df_list.append((df, compare_index_list[i]))

    if os.path.exists(output_path):
        logger.warning(f"{output_path} will be deleted.")
        remove_path(output_path)
    save_excel(output_path, final_result_df_list)
    logger.info(f"The compare results of the multi-ranks are merged and saved in: {output_path}.")


def df_merge(all_result_df_list):
    """
    merge different rank result_df
    """
    if len(all_result_df_list) == 0:
        logger.warning("Nothing to merge.")
        raise CompareException(CompareException.MERGE_COMPARE_RESULT_ERROR)
    if len(all_result_df_list) == 1:
        logger.info("Only one compare result gets merge data.")
    merge_df_base = all_result_df_list[0]
    for sublist in all_result_df_list[1:]:
        for i, sub_df in enumerate(sublist):
            merge_df_base[i] = pd.merge(merge_df_base[i], sub_df, on=CompareConst.NPU_NAME, how='outer')
    for i, value in enumerate(merge_df_base):
        merge_df_base[i] = value.reindex(
            columns=[CompareConst.NPU_NAME] + [col for col in value.columns if col != CompareConst.NPU_NAME])
    return merge_df_base


share_compare_index_list = []


def initialize_compare_index(config):
    global share_compare_index_list
    manager = multiprocessing.Manager()
    share_compare_index_list = manager.list(config.get("compare_index", []))  # 创建共享全局列表


def merge_result(input_dir, output_dir, config_path):
    input_dir = FileChecker(input_dir, FileCheckConst.DIR, FileCheckConst.READ_ABLE).common_check()
    create_directory(output_dir)

    compare_result_path_list = get_result_path(input_dir)   # 获得的input_dir中所有比对结果件的全路径，数量少于2，便提示退出

    config = load_yaml(config_path)
    config = check_config(config)
    api_list = config.get('api')

    # 初始化共享全局变量share_compare_index_list
    initialize_compare_index(config)

    func_args = (compare_result_path_list, api_list)
    all_compare_index_dict_list, all_rank_num_list, all_compare_index_list_list = (
        handle_multi_process(result_process, func_args, multiprocessing.Manager().RLock()))

    generate_merge_result(all_compare_index_dict_list, all_rank_num_list, all_compare_index_list_list, output_dir)



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import subprocess

from msprobe.core.common.log import logger
from msprobe.core.compare.utils import print_compare_ends_info
from msprobe.infer.offline.compare.msquickcmp.adapter_cli.args_adapter import CmpArgsAdapter
from msprobe.infer.offline.compare.msquickcmp.cmp_process import cmp_process
from msprobe.infer.offline.compare.msquickcmp.common.args_check import (
    check_model_path_legality,
    check_output_path_legality,
    check_dict_kind_string,
    check_rank_range_valid,
    check_number_list,
    check_dym_range_string,
    str2bool,
    check_input_data_path
)
from msprobe.infer.offline.compare.msquickcmp.dump.args_adapter import DumpArgsAdapter
from msprobe.infer.offline.compare.msquickcmp.dump.dump_process import dump_process
from msprobe.core.common.file_utils import check_file_or_directory_path

CANN_PATH = os.environ.get('ASCEND_TOOLKIT_HOME', "/usr/local/Ascend/ascend-toolkit/latest")


def _offline_dump_parser(parser):
    parser.add_argument(
        '--model_path',
        required=True,
        dest="model_path",
        type=check_model_path_legality,
        help='The original model .onnx or .om file path'
    )
    parser.add_argument(
        '--input_data',
        default='',
        dest="input_data",
        type=check_input_data_path,
        help='The input data path of the model. Separate multiple inputs with commas(,).'
             ' E.g: input_0.bin,input_1.bin'
    )
    parser.add_argument(
        '-o',
        '--output_path',
        dest="output_path",
        default='./output',
        type=check_output_path_legality,
        help='The output path'
    )
    parser.add_argument(
        '--input_shape',
        type=check_dict_kind_string,
        dest="input_shape",
        default='',
        help="Shape of input shape. Separate multiple nodes with semicolons(;)."
             " E.g: \"input_name1:1,224,224,3;input_name2:3,300\""
    )
    parser.add_argument(
        '--rank',
        type=check_rank_range_valid,
        dest="rank",
        default='0',
        help='Input rank ID [0, 255].'
    )
    parser.add_argument(
        '--dym_shape_range',
        type=check_dym_range_string,
        dest="dym_shape_range",
        default='',
        help="Dynamic shape range using in dynamic model, "
             "using this means ignore input_shape"
             " E.g: \"input_name1:1,3,200~224,224-230;input_name2:1,300\""
    )
    parser.add_argument(
        '--output_size',
        dest="output_size",
        default='',
        help='The size of output. Separate multiple sizes with commas(,). E.g: 10200,34000'
    )
    parser.add_argument(
        '--onnx_fusion_switch',
        dest="onnx_fusion_switch",
        default=True,
        type=str2bool,
        help='Onnxruntime fusion switch, set False for dump complete onnx data when '
             'necessary.Usage: --onnx_fusion_switch False'
    )


def offline_dump_cli(args):
    dump_args = DumpArgsAdapter(
        args.model_path,
        input_data=args.input_data,
        cann_path=CANN_PATH,
        output_path=args.output_path,
        input_shape=args.input_shape, rank=args.rank,
        dym_shape_range=args.dym_shape_range,
        onnx_fusion_switch=args.onnx_fusion_switch,
        output_size=args.output_size
    )
    dump_process(dump_args)


def compare_offline_model_mode(args):
    args = set_args_default(args)
    check_compare_args(args)

    if not args.golden_path:
        logger.error("The following args are required: -gp/--golden_path")
        return
    cmp_args = CmpArgsAdapter(
        args.golden_path,
        args.target_path,
        args.input_data,
        CANN_PATH,
        args.output_path,
        args.input_shape,
        args.rank,
        args.output_size,
        args.dym_shape_range,
        args.onnx_fusion_switch
    )
    cmp_process(cmp_args)
    print_compare_ends_info()


def set_args_default(args):
    if not hasattr(args, 'rank') or not args.rank:
        args.rank = '0'
    if not hasattr(args, 'input_data'):
        args.input_data = ''
    if not hasattr(args, 'input_shape'):
        args.input_shape = ''
    if not hasattr(args, 'output_size'):
        args.output_size = ''
    if not hasattr(args, 'dym_shape_range'):
        args.dym_shape_range = ''
    if not hasattr(args, 'onnx_fusion_switch'):
        args.onnx_fusion_switch = True
    return args


def check_compare_args(args):
    if args.target_path:
        check_file_or_directory_path(args.target_path, False, False, [".om"])
    if args.golden_path:
        check_file_or_directory_path(args.golden_path, False, False, [".onnx", ".om"])
    if args.output_path:
        args.output_path = check_output_path_legality(args.output_path)
    if args.input_data:
        args.input_data = check_input_data_path(args.input_data)
    if args.input_shape:
        args.input_shape = check_dict_kind_string(args.input_shape)
    if args.rank:
        args.rank = check_rank_range_valid(args.rank)
    if args.output_size:
        args.output_size = check_number_list(args.output_size)
    if args.dym_shape_range:
        args.dym_shape_range = check_dym_range_string(args.dym_shape_range)
    if args.onnx_fusion_switch:
        args.onnx_fusion_switch = str2bool(args.onnx_fusion_switch)


def _install_offline_deps_parser(parser):
    parser.add_argument(
        "--no_check",
        dest="no_check",
        action="store_true",
        help="<optional> Whether to skip checking the target website's certificate information "
             "when installing aclruntime and ais_bench poses a certain security risk. "
             "This poses a certain security risk, "
             "and users should use it with caution and bear the consequences themselves.",
        required=False
    )


def install_offline_deps_cli(args):
    offline_extra_install_cmd = [
        "/bin/bash",
        os.path.abspath(os.path.join(os.path.dirname(__file__), "install_aclruntime_aisbench.sh")),
        str(args.no_check)
    ]
    subprocess.run(offline_extra_install_cmd, shell=False)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
import json
import pickle
import time
from msprobe.core.common.file_utils import FileOpen
from msprobe.core.common.const import CompareConst, Const
from msprobe.core.common.log import logger
from msprobe.core.common.exceptions import MsprobeException
from msprobe.core.compare.utils import check_and_return_dir_contents
from msprobe.core.common.utils import CompareException


def load_json_file(file_path):
    """
    加载json文件
    """
    try:
        with FileOpen(file_path, 'r') as f:
            file_dict = json.load(f)
            if not isinstance(file_dict, dict):
                return {}
            return file_dict
    except json.JSONDecodeError:
        return {}


def load_data_json_file(file_path):
    """
    加载dump.json中的data字段
    """
    return load_json_file(file_path).get(GraphConst.DATA_KEY, {})


def str2float(percentage_str):
    """
    百分比字符串转换转换为浮点型
    Args:
        percentage_str: '0.00%', '23.4%'
    Returns: float 0.00, 0.234
    """
    try:
        percentage_str = percentage_str.strip('%')
        return float(percentage_str) / 100
    except (ValueError, AttributeError):
        return 0


def get_step_or_rank_int(x: str, is_rank=False):
    """
    获取字符串rank{int}或者step{int}中的int值，如果x=rank或step，返回0
    """
    if x in [Const.RANK, Const.STEP] or x is None:
        return 0
    description = Const.RANK if is_rank else Const.STEP
    try:
        x_int = int(x.replace(Const.RANK, "")) if is_rank else int(x.replace(Const.STEP, ""))
    except Exception as e:
        logger.error(f'The folder name format is incorrect, expected {description}+number, such as rank0, step1, etc.')
        raise RuntimeError from e
    return x_int


def check_directory_content(input_path):
    """
    检查input_path内容, 是否全是step{数字}命名的文件夹(例如step0), 或者全是rank{数字}命名的文件夹(例如rank0), 或者全是文件
    """
    contents = os.listdir(input_path)
    if not contents:
        raise ValueError(f'The path {input_path} is empty.')

    # 真实数据dump会有dump_tensor_data文件夹
    if os.path.exists(os.path.join(input_path, Const.DUMP_TENSOR_DATA)):
        return GraphConst.FILES

    # 检查是否全是文件
    if all(os.path.isfile(os.path.join(input_path, item)) for item in contents):
        return GraphConst.FILES

    # 单卡只有一个rank文件夹
    if contents == [Const.RANK]:
        return GraphConst.RANKS

    rank_pattern = re.compile(r'^rank\d+$')
    step_pattern = re.compile(r'^step\d+$')

    rank_all = True
    step_all = True

    for item in contents:
        item_path = os.path.join(input_path, item)
        if not os.path.isdir(item_path):
            continue
        if not rank_pattern.match(item):
            rank_all = False
        if not step_pattern.match(item):
            step_all = False

    if rank_all:
        return GraphConst.RANKS
    if step_all:
        return GraphConst.STEPS

    raise ValueError("The input path content does not conform to the expected naming convention. "
                     "It is expected to be all step{number} named folders (such as step0), "
                     "all rank{number} named folders (such as rank0), or all files.")


def extract_rank_number(rank_str):
    try:
        return int(rank_str[4:])
    except ValueError:
        return 0


def sort_rank_number_strings(rank_number_strings):
    sorted_list = sorted(rank_number_strings, key=extract_rank_number)
    return sorted_list


def load_parallel_param(args):
    param_t = [args.rank_size, args.tp, args.pp]
    if any(not x for x in param_t):
        logger.error('In the graph merge task, parameters "tp/pp/rank_size" is required!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    if not args.golden_path and len(args.rank_size) != 1:
        logger.error('In the graph merge build task, '
                     'the number of parameters "rank_size" to be filled in is either 1!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    if args.golden_path and len(args.rank_size) != 2:
        logger.error('In the graph merge compare task, '
                     'the number of parameters "rank_size" to be filled in is either 2!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    if len(args.rank_size) == 1:
        if any(len(x) != 1 for x in param_t):
            logger.error('In the graph merge build task, '
                         'the number of parameters "tp/pp/rank_size" to be filled in is either 1!')
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

        param_n = ParallelParam(args.rank_size[0], args.tp[0], args.pp[0], args.vpp[0] if args.vpp else 1,
                                args.order[0] if args.order else 'tp-cp-ep-dp-pp')
        return (param_n,)
    else:
        if any(len(x) != 2 for x in param_t):
            logger.error('In the graph merge compare task, '
                         'the number of parameters "tp/pp/rank_size" to be filled in is either 2!')
            raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

        param_n = ParallelParam(args.rank_size[0], args.tp[0], args.pp[0], args.vpp[0] if len(args.vpp) == 2 else 1,
                                args.order[0] if len(args.order) == 2 else 'tp-cp-ep-dp-pp')
        param_b = ParallelParam(args.rank_size[1], args.tp[1], args.pp[1], args.vpp[1] if len(args.vpp) == 2 else 1,
                                args.order[1] if len(args.order) == 2 else 'tp-cp-ep-dp-pp')
        return param_n, param_b


def validate_parallel_param(parallel_param, dump_path, log_prefix='[NPU]'):
    pattern = re.compile(r'^[a-z\-]+$')
    params = [parallel_param.tp, parallel_param.pp, parallel_param.rank_size, parallel_param.vpp]
    ranks = check_and_return_dir_contents(dump_path, Const.RANK)
    if len(ranks) != parallel_param.rank_size:
        logger.error(f'{log_prefix} The parallel param "rank_size" error, '
                     f'you set {parallel_param.rank_size} but expected {len(ranks)}.')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if any(x is None for x in params):
        logger.error(f'{log_prefix} The parallel params "tp/pp/rank_size" must not be null!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if any(isinstance(x, bool) for x in params):
        logger.error(f'{log_prefix} The parallel params "tp/pp/vpp/rank_size" must not be bool!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if any(not isinstance(x, int) for x in params):
        logger.error(f'{log_prefix} The parallel params "tp/pp/vpp/rank_size" must be int!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if any(x <= 0 for x in params):
        logger.error(f'{log_prefix} The parallel params "tp/pp/vpp/rank_size" must be greater than 0!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if parallel_param.tp > parallel_param.rank_size:
        logger.error(f'{log_prefix} The parallel param "tp" must be less than or equal to "rank_size"!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if parallel_param.pp > parallel_param.rank_size:
        logger.error(f'{log_prefix} The parallel param "pp" must be less than or equal to "rank_size"!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if parallel_param.rank_size % parallel_param.tp != 0:
        logger.error(f'{log_prefix} The parallel param "rank_size" must be divisible by "tp"!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if parallel_param.rank_size % parallel_param.pp != 0:
        logger.error(f'{log_prefix} The parallel param "rank_size" must be divisible by "pp"!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if parallel_param.tp * parallel_param.pp > parallel_param.rank_size:
        logger.error(f'{log_prefix} The parallel params "tp * pp" must be less than or equal to "rank_size"!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if parallel_param.vpp > 1 and parallel_param.pp < 2:
        logger.error(f'{log_prefix} When configuring the parallel param "vpp", the "pp" param must be greater than 1!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if not isinstance(parallel_param.order, str):
        logger.error(f'{log_prefix} The parallel params "order" must be of string type!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if not pattern.match(parallel_param.order):
        logger.error(f'{log_prefix} The parallel params "order" must consist only of lowercase letters and "-"!')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)


class ParallelParam:
    def __init__(self, rank_size, tp, pp, vpp=1, order='tp-cp-ep-dp-pp'):
        self.rank_size = rank_size
        self.tp = tp
        self.pp = pp
        self.vpp = vpp
        self.order = order


class ToolTip:
    MAX_DIFF = 'NPU与标杆API统计信息比对，最大值的差值'
    MIN_DIFF = 'NPU与标杆API统计信息比对，最小值的差值'
    MEAN_DIFF = 'NPU与标杆API统计信息比对，平均值的差值'
    NORM_DIFF = 'NPU与标杆API统计信息比对，2范数（平方根）的差值'
    MD5 = '数据MD5信息，用于比较两个数据信息是否完全一致'
    ONE_THOUSANDTH_ERR_RATIO = 'Tensor中的元素逐个与对应的标杆数据对比，相对误差小于千分之一的比例占总元素个数的比例，比例越接近1越好'
    FIVE_THOUSANDTHS_ERR_RATIO = 'Tensor中的元素逐个与对应的标杆数据对比，相对误差小于千分之五的比例占总元素个数的比例，比例越接近1越好'
    COSINE = (
        '通过计算两个向量的余弦值来判断其相似度，数值越接近于1说明计算出的两个张量越相似，实际可接受阈值为大于0.99。'
        '在计算中可能会存在nan，主要由于可能会出现其中一个向量为0'
    )
    MAX_ABS_ERR = '当最大绝对误差越接近0表示其计算的误差越小，实际可接受阈值为小于0.001'
    MAX_RELATIVE_ERR = (
        '当最大相对误差越接近0表示其计算的误差越小。'
        '当dump数据中存在0或Nan时，比对结果中最大相对误差则出现inf或Nan的情况，属于正常现象'
    )


class GraphConst:
    CONSTRUCT_FILE = 'construct.json'
    DUMP_FILE = 'dump.json'
    STACK_FILE = 'stack.json'
    ERROR_KEY = 'error_key'
    SUMMARY_COMPARE = 0
    MD5_COMPARE = 1
    REAL_DATA_COMPARE = 2
    STRUCTURE_COMPARE = 3
    JSON_NPU_KEY = 'NPU'
    JSON_BENCH_KEY = 'Bench'
    JSON_TIP_KEY = 'ToolTip'
    JSON_ROOT_KEY = 'root'
    JSON_NODE_KEY = 'node'
    JSON_DATA_KEY = 'dump_data_dir'
    JSON_TASK_KEY = 'task'
    DATA_KEY = 'data'
    ROUND_TH = 6
    JSON_INDEX_KEY = 'precision_index'
    MATCHED_DISTRIBUTED = 'matched_distributed'
    OVERFLOW_LEVEL = 'overflow_level'
    MAX_INDEX_KEY = 1
    MIN_INDEX_KEY = 0
    INPUT = '.input.'
    OUTPUT = '.output.'
    STR_MAX_LEN = 50
    MD5_INDEX_LIST = CompareConst.MD5_COMPARE_INDEX + [CompareConst.REQ_GRAD_CONSIST, CompareConst.RESULT,
                                                       CompareConst.ERROR_MESSAGE]
    REAL_DATA_INDEX_LIST = CompareConst.ALL_COMPARE_INDEX + [CompareConst.REQ_GRAD_CONSIST, CompareConst.RESULT,
                                                             CompareConst.ERROR_MESSAGE]
    SUMMARY_INDEX_LIST = CompareConst.SUMMARY_COMPARE_INDEX + [CompareConst.REQ_GRAD_CONSIST, CompareConst.RESULT,
                                                               CompareConst.ERROR_MESSAGE]
    APIS_BETWEEN_MODULES = 'Apis_Between_Modules'
    APIS_BETWEEN_MODULES_ALL_RANKS = 'Apis_Between_Modules_All_Ranks'
    NULL = 'null'
    NONE = 'None'
    VALUE = 'value'
    DESCRIPTION = 'description'
    ACCURACY_LEVEL = 'accuracy_level'
    COLORS = 'Colors'
    MICRO_STEPS = 'MicroSteps'
    OVERFLOW_CHECK = 'OverflowCheck'
    UNMATCHED = 'unmatched node'

    DUMP_MODE_TO_GRAPHCOMPARE_MODE_MAPPING = {
        Const.ALL: REAL_DATA_COMPARE,
        Const.SUMMARY: SUMMARY_COMPARE,
        Const.MD5: MD5_COMPARE,
        Const.STRUCTURE: STRUCTURE_COMPARE
    }

    GRAPHCOMPARE_MODE_TO_DUMP_MODE_TO_MAPPING = {
        REAL_DATA_COMPARE: Const.ALL,
        SUMMARY_COMPARE: Const.SUMMARY,
        MD5_COMPARE: Const.MD5,
        STRUCTURE_COMPARE: Const.STRUCTURE
    }

    COMPARE_INDICATOR_TO_PRECISION_INDEX_MAPPING = {
        CompareConst.PASS: 0,
        CompareConst.WARNING: 0.5,
        CompareConst.ERROR: 1
    }

    RANKS = 'ranks'
    STEPS = 'steps'
    FILES = 'files'

    SRC = 'src'
    DST = 'dst'

    BATCH_P2P = 'batch_isend_irecv'
    OP = 'op'
    PEER = 'peer'
    GROUP_ID = 'group_id'

    UNCERTAINTY_THRESHOLD = 1e-6
    REDUCE_OPERATIONS = ['reduce_scatter', 'all_reduce']

    IGNORE_PRECISION_INDEX = {'empty', 'empty_like', 'empty_with_format', 'new_empty_strided', 'new_empty',
                              'empty_strided'}
    VPP_CHUNK_0 = '0'

    PBAR_TOTAL = 100
    BAR_FORMAT = "{l_bar}{bar}| {percentage:3.0f}% [{elapsed}<{remaining}, {rate_fmt}] "
    PBAR_DESC_PREFIX = '[Graph] Overall Progress'
    BUILD_STAGES_TOTAL = 2
    COMPARE_STAGES_TOTAL = 5

    INTERSECTION = 'intersection'
    UNION = 'union'


def is_serializable(obj):
    """
    Check if an object is serializable
    """
    try:
        pickle.dumps(obj)
        return True
    except (pickle.PicklingError, AttributeError, TypeError):
        return False
    except Exception as e:
        logger.error('Unexpected error occurred while pickling obj.')
        raise RuntimeError('Unexpected error occurred while pickling obj.') from e


class SerializableArgs:
    def __init__(self, args):
        for k, v in vars(args).items():
            if is_serializable(v):
                setattr(self, k, v)


def calculate_list(path_n, path_b, prefix_str=Const.RANK, mode=GraphConst.INTERSECTION):
    """
    计算列表，可取交集、并集
    """
    if prefix_str not in [Const.RANK, Const.STEP]:
        logger.error(f'Parameter "prefix_str" is expected to be {Const.RANK} or {Const.STEP}, '
                     f'but in reality it is {prefix_str}.')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)
    if mode not in [GraphConst.INTERSECTION, GraphConst.UNION]:
        logger.error(f'Parameter "mode" is expected to be {GraphConst.INTERSECTION} or {GraphConst.UNION}, '
                     f'but in reality it is {mode}.')
        raise MsprobeException(MsprobeException.INVALID_PARAM_ERROR)

    npu_list = check_and_return_dir_contents(path_n, prefix_str)
    bench_list = check_and_return_dir_contents(path_b, prefix_str)

    if not npu_list:
        logger.error(f'Cannot get {prefix_str} in {path_n}.')
        raise CompareException(CompareException.INVALID_PATH_ERROR)

    if npu_list != bench_list:
        result_list = npu_list
        if mode == GraphConst.INTERSECTION:
            result_list = sort_rank_number_strings(list(set(npu_list) & set(bench_list)))
        elif mode == GraphConst.UNION:
            result_list = sort_rank_number_strings(list(set(npu_list) | set(bench_list)))

        if not result_list:
            logger.error(f'Two {prefix_str}s cannot obtain {mode}.')
            raise CompareException(CompareException.INVALID_PATH_ERROR)
        return result_list
    return npu_list


POST_DB_PROCESS = 0


class ProgressInfo:
    """
    前端监测此类属性获取进度信息
    """
    current_progress = 0
    process_running = True
    error_msg = []
    info_dict = {}
    print_progress_log = False

    @classmethod
    def reset(cls):
        cls.current_progress = 0
        cls.process_running = True
        cls.error_msg = []

    @classmethod
    def update_current_progress(cls, value):
        if cls.current_progress != value:
            cls.current_progress = value
            cls.print_info()

    @classmethod
    def update_process_running(cls, value):
        if cls.process_running != value:
            cls.process_running = value
            cls.print_info()

    @classmethod
    def update_error_msg(cls, value):
        cls.error_msg.append(value)
        cls.print_info()

    @classmethod
    def print_info(cls):
        if cls.print_progress_log:
            cls.info_dict["current_progress"] = cls.current_progress
            cls.info_dict["process_running"] = cls.process_running
            cls.info_dict["error_msg"] = cls.error_msg
            logger.info(f'PROGRESS: {json.dumps(cls.info_dict)}')


def get_log_msg_wrapper(fn):
    def decorated(self, msg):
        ProgressInfo.update_error_msg(msg)
        return fn(self, msg)

    return decorated


def monitor_progress(pbar_info, pbar, all_task_ids, is_parallel_merge=False):
    """监测所有任务进度（包括等待中的），取平均值更新总进度"""
    ProgressInfo.total_stage = pbar_info.stage_total
    task_progress = pbar_info.progress_dict
    for task_id in all_task_ids:
        task_progress[task_id] = 0
    # 图合并模式，总进度卡99%，剩余的1%给db后处理函数; 否则总进度卡98%，剩余的1%给db后处理函数，1%给分布式节点分析关联
    global POST_DB_PROCESS
    POST_DB_PROCESS = pbar_info.total - pbar_info.step_total if is_parallel_merge \
        else pbar_info.total - 2 * pbar_info.step_total
    all_complete = False

    while not all_complete:
        if pbar_info.stop_monitor:
            break
        if pbar_info.wait_monitor:
            logger.info('Detected wait_monitor=True, the progress monitoring thread has entered a paused state...')
            while pbar_info.wait_monitor:
                time.sleep(0.1)
                if pbar_info.stop_monitor:
                    break
            if pbar_info.continue_monitor:
                logger.info('Detected continue_monitor=True, the progress monitoring thread resumes running.')
        # 总进度取所有进程进度的平均值，但不超过数据库后处理进度
        current_progress = min(sum(task_progress.values()) // len(task_progress), POST_DB_PROCESS)
        ProgressInfo.update_current_progress(current_progress // pbar_info.step_total)
        pbar.n = current_progress
        pbar.refresh()
        # 检查是否所有任务都已完成: 数据库后处理是否完成
        all_complete = POST_DB_PROCESS == pbar_info.total
        # 降低检查频率
        time.sleep(0.1)

    post_process_pbar(pbar_info, pbar)


def update_shared_dict(shared_dict, key, value, update_all=False):
    """
    进程函数，根据键添加或更新共享字典
    """
    if update_all:
        for k in shared_dict.keys():
            shared_dict[k] += value
    else:
        shared_dict[key] = shared_dict.get(key, 0) + value


def update_pbar_info(pbar_info, current_data, total_data, step_ratio=0.01, update_all=False):
    """
    带步长控制的进度条更新函数
    :param pbar_info: 进度条信息对象
    :param current_data: 当前进度位置
    :param total_data: 总数据量
    :param step_ratio: 更新步长比例（默认每1%更新一次）
    :param update_all
    """
    step = max(1, int(total_data * step_ratio))

    # 仅在最后一次/达到步长时执行更新
    if current_data != total_data and current_data % step != 0:
        return

    progress = (current_data * pbar_info.total // total_data) // pbar_info.stage_total
    progress += (pbar_info.stage_progress * (pbar_info.current_stage_dict.get(pbar_info.task_id) - 1))
    if pbar_info.pbar is not None:
        if progress == pbar_info.total:
            # 卡99%，剩余的1%给db后处理函数
            progress -= 1
        pbar_info.pbar.n = progress
        pbar_info.pbar.refresh()
        ProgressInfo.update_current_progress(progress)
    elif pbar_info.progress_dict is not None:
        if update_all:
            for task_id in pbar_info.progress_dict.keys():
                pbar_info.progress_dict[task_id] = progress
        else:
            pbar_info.progress_dict[pbar_info.task_id] = progress


def post_process_db_pbar(pbar_info, is_parallel_merge=False):
    if pbar_info.progress_dict is not None:
        global POST_DB_PROCESS
        POST_DB_PROCESS = pbar_info.total if is_parallel_merge else pbar_info.total - 1
    elif pbar_info.pbar is not None:
        pbar_info.pbar.n = pbar_info.total
        pbar_info.pbar.refresh()
        ProgressInfo.update_current_progress(pbar_info.total)


def post_process_pbar(pbar_info, pbar):
    pbar.n = pbar_info.total
    ProgressInfo.update_current_progress(pbar_info.total // pbar_info.step_total)
    pbar.set_description("[Graph] All tasks completed")
    pbar.refresh()


def update_progress_info(proc, progress_info: ProgressInfo):
    """
    通过管道获取日志，解析日志进度信息，更新进度类，供前端调用
    """
    try:
        while True:
            line = proc.stdout.readline()
            if line == '':
                if proc.poll() is not None:
                    break
            else:
                line = line.strip()
                if 'PROGRESS: ' in line:
                    line = line.split('PROGRESS: ')[-1]
                    info_dict = json.loads(line)
                    progress_info.current_progress = info_dict.get('current_progress')
                    progress_info.process_running = info_dict.get('process_running')
                    progress_info.error_msg = info_dict.get('error_msg')
            time.sleep(0.1)
    finally:
        proc.stdout.close()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import importlib
import inspect

from msprobe.core.common.file_utils import load_yaml, check_link
from msprobe.core.common.log import logger


def get_ops():
    cur_path = os.path.dirname(os.path.realpath(__file__))
    yaml_path = os.path.join(cur_path, "support_wrap_ops.yaml")
    ops = load_yaml(yaml_path)
    wrap_functional = ops.get('functional')
    wrap_tensor = ops.get('tensor')
    wrap_torch = ops.get('torch')
    wrap_npu_ops = ops.get('torch_npu')
    return set(wrap_functional) | set(wrap_tensor) | set(wrap_torch) | set(wrap_npu_ops)


def dynamic_import_op(package, white_list):
    package_name = package.__name__
    ops = {}
    ops_dir, _ = os.path.split(package.__file__)
    check_link(ops_dir)
    for file_name in os.listdir(ops_dir):
        if file_name in white_list:
            sub_module_name = file_name[:-3]
            module_name = f"{package_name}.{sub_module_name}"
            try:
                module = importlib.import_module(module_name)
            except Exception as e:
                logger.warning(f"import {module_name} failed!")
                continue

            func_members = inspect.getmembers(module, inspect.isfunction)
            for func_member in func_members:
                func_name, func = func_member[0], func_member[1]
                ops[f"{sub_module_name}.{func_name}"] = func
    return ops



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import re
import hashlib

from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.log import logger
from msprobe.core.common.const import Const


def merge_keys(dir_0, dir_1):
    output_list = list(dir_0.keys())
    output_list.extend(list(dir_1.keys()))
    return set(output_list)


def compare_dict(bench_dict, cmp_dict):
    result = []
    for key in set(bench_dict.keys()) | set(cmp_dict.keys()):
        if key in bench_dict and key in cmp_dict:
            if bench_dict[key] != cmp_dict[key]:
                result.append(f"{key}: {bench_dict[key]} -> {cmp_dict[key]}")
        elif key in bench_dict:
            result.append(f"{key}: [deleted] -> {bench_dict[key]}")
        else:
            result.append(f"{key}: [added] -> {cmp_dict[key]}")
    return result


def config_checking_print(msg):
    logger.info(f"[config checking log] {msg}")


def tensor_to_hash(tensor):
    """Compute the hash value of a tensor"""
    tensor_bytes = tensor.clone().detach().cpu().numpy().tobytes()
    return bytes_hash(tensor_bytes)


def get_tensor_features(tensor):
    features = {
        "max": FmkAdp.tensor_max(tensor),
        "min": FmkAdp.tensor_min(tensor),
        "mean": FmkAdp.tensor_mean(tensor),
        "norm": FmkAdp.tensor_norm(tensor),
    }

    return features


def compare_dicts(dict1, dict2, path=''):
    deleted = []
    added = []
    changed = []
    result = {}

    for key in dict1:
        if key not in dict2:
            deleted.append(f"[Deleted]: {path + key}")
            result[key] = "[deleted]"
        else:
            if isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
                sub_deleted, sub_added, sub_changed, sub_result = compare_dicts(
                    dict1[key], dict2[key], path + key + '/')
                deleted.extend(sub_deleted)
                added.extend(sub_added)
                changed.extend(sub_changed)
                if sub_result:
                    result[key] = sub_result
            elif dict1[key] != dict2[key]:
                changed.append(f"[Changed]: {path + key} : {dict1[key]} -> {dict2[key]}")
                result[key] = f"[changed]: {dict1[key]} -> {dict2[key]}"
    for key in dict2:
        if key not in dict1:
            added.append(f"[Added]: {path + key}")
            result[key] = "[added]"
    return deleted, added, changed, result


def bytes_hash(obj: bytes):
    hex_dig = hashlib.sha256(obj).hexdigest()
    short_hash = int(hex_dig, 16) % (2 ** 16)
    return short_hash


def update_dict(ori_dict, new_dict):
    for key, value in new_dict.items():
        if key in ori_dict and ori_dict[key] != value:
            if "values" in ori_dict.keys():
                ori_dict[key]["values"].append(new_dict[key])
            else:
                ori_dict[key] = {"description": "duplicate_value", "values": [ori_dict[key], new_dict[key]]}
        else:
            ori_dict[key] = value


def process_pass_check(data):
    if Const.CONFIG_CHECK_ERROR in data:
        return Const.CONFIG_CHECK_ERROR
    elif Const.CONFIG_CHECK_WARNING in data:
        return Const.CONFIG_CHECK_WARNING
    else:
        return Const.CONFIG_CHECK_PASS



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.file_utils import load_yaml


class ModeConfig:
    def __init__(self, **kwargs):
        self.stack_mode = kwargs.get('stack_mode', False)
        self.fuzzy_match = kwargs.get('fuzzy_match', False)
        self.dump_mode = kwargs.get('dump_mode', Const.SUMMARY)
        self.first_diff_analyze = kwargs.get('first_diff_analyze', False)
        self.diff_analyze = kwargs.get('diff_analyze', False)
        self.compared_file_type = kwargs.get('compared_file_type', Const.DUMP_JSON_FILE)


class MappingConfig:
    def __init__(self, cell_mapping=None, api_mapping=None, data_mapping=None):
        self.cell_mapping = cell_mapping
        self.api_mapping = api_mapping
        self.data_mapping = data_mapping


class MappingDict:
    def __init__(self, mapping_config: MappingConfig):
        self.cell_mapping_dict = self.load_mapping_file(mapping_config.cell_mapping)
        self.api_mapping_dict = self.load_mapping_file(mapping_config.api_mapping)
        if mapping_config.api_mapping is not None:
            self.ms_to_pt_mapping = self.load_internal_api()
        self.data_mapping_dict = self.init_data_mapping(mapping_config.data_mapping)

    @staticmethod
    def load_internal_api():
        cur_path = os.path.dirname(os.path.realpath(__file__))
        yaml_path = os.path.abspath(os.path.join(cur_path, CompareConst.INTERNAL_API_MAPPING_FILE))
        return load_yaml(yaml_path)

    @staticmethod
    def load_mapping_file(mapping_file):
        if isinstance(mapping_file, str):
            mapping_dict = load_yaml(mapping_file)
        else:
            mapping_dict = {}
        return mapping_dict

    def init_data_mapping(self, data_mapping):
        """
        初始化data_mapping_dict
        """
        if isinstance(data_mapping, str) or data_mapping is None:
            data_mapping_dict = self.load_mapping_file(data_mapping)
        elif isinstance(data_mapping, dict):
            data_mapping_dict = data_mapping
        else:
            raise TypeError(f"The type of parameter `data_mapping` must be dict, str or None, but got "
                            f"{type(data_mapping)}")
        return data_mapping_dict



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
# 前向函数声明对比
标杆实现:fusion_attention_forward: q, k, v, drop_mask, atten_mask, pse, scale, keep_prob
融合算子:npu_fusion_attention_forward: query, key, value, head_num, input_layout, *, pse=None, padding_mask=None,
                                      atten_mask=None, scale=1.0, keep_prob=1.0, pre_tockens=2147483647,
                                      next_tockens=2147483647, inner_precise=0, prefix=None, sparse_mode=0,
                                      gen_mask_parallel=True, sync=False

# 反向函数声明对比
标杆实现:fusion_attention_backward: dx, q, k, v, softmax_res, drop_mask, pse, scale, keep_prob
融合算子:npu_fusion_attention_backward: query, key, value, dy, head_num, input_layout, *, pse=None, padding_mask=None,
                                       atten_mask=None, softmax_max=None, softmax_sum=None, softmax_in=None,
                                       attention_in=None, scale_value=1.0, keep_prob=1.0, pre_tockens=2147483647,
                                       next_tockens=2147483647, inner_precise=0, seed=0, offset=0,
                                       numels=0, prefix=None, sparse_mode=0, gen_mask_parallel=True, sync=False
"""

from collections import namedtuple
import torch
import numpy as np
from einops import rearrange

try:
    import torch_npu
except ImportError:
    is_gpu = True
    try:
        # flash_attn为gpu的fa三方库
        from flash_attn import flash_attn_func
    except ImportError:
        # 如果为cpu的ut环境，则不做任何处理
        pass
else:
    is_gpu = False

from msprobe.pytorch.common.utils import logger
from msprobe.core.common.const import Const, CompareConst

GTYPE = torch.float64  # arm host必须选择float64，x86环境选择float32即可，64也行。arm计算很慢，s=8k的场景建议使用x86
SOFTMAX_BUILD_MODE = "QKV"  # "MAX_SUM"


FaForwardParams = namedtuple("FaForwardParams",
                            ["q", "k", "v", "drop_mask", "atten_mask", "pse", "scale", "keep_prob"])
FaBackwardParams = namedtuple("FaBackwardParams",
                            ["dx", "q", "k", "v", "softmax_res", "drop_mask", "pse", "scale", "keep_prob"])
RebuildSoftmaxParams = namedtuple("RebuildSoftmaxParams",
                                ["q", "k", "atten_mask", "pse", "scale", "softmax_max", "softmax_sum"])


def softmax_forward(x):
    x_max = torch.max(x, dim=-1, keepdims=True)[0]
    x_sub = x.sub(x_max)
    y = torch.exp(x_sub)
    x_sum = y.sum(dim=-1, keepdims=True)
    res = y.div(x_sum)
    return res, x_max, x_sum


def softmax_grad(dp, softmax_res):
    muls = dp * softmax_res
    muls_r = muls.sum(dim=-1, keepdims=True)
    sub_r = dp - muls_r
    res = sub_r * softmax_res
    return res


def broadcast_kv(num_heads, num_kv_heads, kv_tensor, dtype):
    if num_kv_heads == 0 or num_kv_heads > num_heads:
        raise ValueError(f"num_kv_heads must be non-zero and bigger than num_heads.")

    factor = num_heads // num_kv_heads
    kv_shape = kv_tensor.shape
    b = kv_shape[0]
    s = kv_shape[2]
    d = kv_shape[3]
    kv_res = torch.zeros([b, num_heads, s, d]).to(dtype)
    for i in range(num_heads):
        j = i // factor
        kv_res[:, i:i + 1, :, :] = kv_tensor[:, j:j + 1, :, :]
    return kv_res


def calculate_qk(q, k, atten_mask, pse, scale):
    if pse is None or len(pse.shape) == 0:
        qk = torch.matmul(q, k.permute(0, 1, 3, 2)).mul(scale)
    else:
        qk = (torch.matmul(q, k.permute(0, 1, 3, 2)) + pse).mul(scale)
    if atten_mask is None or len(atten_mask.shape) == 0:
        return qk
    else:
        qk = qk + atten_mask.bool() * (-40000.0)  # -10000
    return qk


def fusion_attention_forward(forward_params):
    q = forward_params.q
    k = forward_params.k
    v = forward_params.v
    drop_mask = forward_params.drop_mask
    atten_mask = forward_params.atten_mask
    pse = forward_params.pse
    scale = forward_params.scale
    keep_prob = forward_params.keep_prob

    # 除零风险拦截：keep_prob 为 0 时会导致除零错误
    if keep_prob == 0:
        raise ValueError("fusion_attention_forward: keep_prob cannot be zero to avoid division by zero.")


    qk = calculate_qk(q, k, atten_mask, pse, scale)
    softmax_res, softmax_max, softmax_sum = softmax_forward(qk)
    if drop_mask is None or len(drop_mask.shape) == 0:
        drop_res = softmax_res
    else:
        drop_res = softmax_res * drop_mask * (1.0 / keep_prob)
    y = torch.matmul(drop_res, v)
    return y, softmax_max, softmax_sum


def fusion_attention_backward(backward_params):
    dx = backward_params.dx
    q = backward_params.q
    k = backward_params.k
    v = backward_params.v
    softmax_res = backward_params.softmax_res
    drop_mask = backward_params.drop_mask
    pse = backward_params.pse
    scale = backward_params.scale
    keep_prob = backward_params.keep_prob

    # 除零风险拦截：keep_prob 为 0 时会导致除零错误
    if keep_prob == 0:
        raise ValueError("fusion_attention_backward: keep_prob cannot be zero to avoid division by zero.")

    dp = torch.matmul(dx, v.permute(0, 1, 3, 2))
    if drop_mask is None or len(drop_mask.shape) == 0:
        drop_res = softmax_res.permute(0, 1, 3, 2)
        dp_drop = dp
    else:
        drop_res = softmax_res.mul(drop_mask).mul(1.0 / keep_prob).permute(0, 1, 3, 2)
        dp_drop = dp * drop_mask * (1.0 / keep_prob)
    dv = torch.matmul(drop_res, dx)
    softmax_grad_res = (softmax_grad(dp_drop, softmax_res) * scale)
    dq = torch.matmul(softmax_grad_res, k)
    dk = torch.matmul(softmax_grad_res.permute(0, 1, 3, 2), q)
    return dq, dk, dv


def parse_bsnd_args(query, key, head_num, input_layout):
    supported_input_layout = ["BSH", "SBH", "BSND", "BNSD", "TND"]
    b, s1, s2, n1, n2, d, h1, h2 = None, None, None, head_num, None, None, None, None

    if not isinstance(input_layout, str) or input_layout not in supported_input_layout:
        raise ValueError(f"Invalid input_layout arg which must be one of {supported_input_layout}.")

    if input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")
    try:
        if input_layout == "BSH":
            b, s1, h1 = query.shape
            _, s2, h2 = key.shape
            if n1 == 0:
                raise ValueError("parse_bsnd_args: head_num (n1) cannot be zero to avoid division by zero.")
            d = h1 // n1
            if d == 0:
                raise ValueError("parse_bsnd_args: computed head dimension (d) is zero, division by zero risk.")
            n2 = h2 // d
        elif input_layout == "SBH":
            s1, b, h1 = query.shape
            s2, _, h2 = key.shape
            if n1 == 0:
                raise ValueError("parse_bsnd_args: head_num (n1) cannot be zero to avoid division by zero.")
            d = h1 // n1
            if d == 0:
                raise ValueError("parse_bsnd_args: computed head dimension (d) is zero, division by zero risk.")
            n2 = h2 // d
        elif input_layout == "BSND":
            b, s1, n1, d = query.shape
            _, s2, n2, _ = key.shape
            h1 = n1 * d
            h2 = n2 * d
            if d == 0:
                raise ValueError("parse_bsnd_args: head dimension (d) is zero, division by zero risk.")
        elif input_layout == "BNSD":
            b, n1, s1, d = query.shape
            _, n2, s2, _ = key.shape
            h1 = n1 * d
            h2 = n2 * d
            if d == 0:
                raise ValueError("parse_bsnd_args: head dimension (d) is zero, division by zero risk.")
    except Exception as e:
        raise ValueError(f"query.shape: {query.shape}, key.shape: {key.shape}, parse_bsnd_args error: {e}") from e

    if d == 0:
        raise ValueError(f"Value d must be non-zero.")
    _dtype = query.dtype
    ret = (b, s1, s2, n1, n2, d, h1, h2, _dtype)
    return ret


def convert_from_bnsd(_input, input_layout):
    """
    transform qkv from bnsd to input_layout.
    B: batch_size
    S: sequence_length
    N: num_heads
    D: head_dim
    Args:
       _input (torch.Tensor): tensor of shape (B,N,S,D)
        input_layout (str): "BSH" or "SBH" or "BSND" or "BNSD" or "TND"
    Returns:
        tensor of shape (B,N,S,D) or (B,S,N,D) or (S,B,H) or (B,S,H)
    """
    if input_layout == "BSH":
        # (B,N,S,D)=>(B,S,N*D)
        out = rearrange(_input, 'b n s d -> b s (n d)').contiguous()
    elif input_layout == "SBH":
        # (B,N,S,D)=>(S,B,N*D)
        out = rearrange(_input, 'b n s d -> s b (n d)').contiguous()
    elif input_layout == "BSND":
        # (B,N,S,D)=>(B,S,N,D)
        out = rearrange(_input, 'b n s d -> b s n d').contiguous()
    elif input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")
    else:
        out = _input
    return out


def convert_to_bnsd(_input, n, input_layout):
    """
    transform qkv from input_layout to bnsd.
    B: batch_size
    S: sequence_length
    N: num_heads
    D: head_dim
    Args:
        _input (torch.Tensor): tensor of shape (B,N,S,D) or (B,S,N,D) or (S,B,H) or (B,S,H)
        n (int): num_heads
        input_layout (str):"BSH" or "SBH" or "BSND" or "BNSD" or "TND"
    Returns:
        tensor of shape (B,N,S,D)
    """
    if input_layout == "BSH":
        # (B,S,N*D)=>(B,N,S,D)
        out = rearrange(_input, 'b s (n d) -> b n s d', n=n)
    elif input_layout == "SBH":
        # (S,B,N*D)=>(B,N,S,D)
        out = rearrange(_input, 's b (n d) -> b n s d', n=n)
    elif input_layout == "BSND":
        # (B,S,N,D)=>(B,N,S,D)
        out = rearrange(_input, 'b s n d -> b n s d', n=n)
    elif input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")
    else:
        out = _input
    if out.dim() != 4:
        raise ValueError(f"convert qkv format failed with input_layout {input_layout}.")
    return out.to(GTYPE)


def convert_from_bsnd(_input, input_layout):
    """
    transform qkv from bsnd to input_layout.
    B: batch_size
    S: sequence_length
    N: num_heads
    D: head_dim
    Args:
       _input (torch.Tensor): tensor of shape (B,S,N,D)
        input_layout (str): "BSH" or "SBH" or "BSND" or "BNSD" or "TND"
    Returns:
        tensor of shape (B,N,S,D) or (B,S,N,D) or (S,B,H) or (B,S,H)
    """
    if input_layout == "BSH":
        # (B,S,N,D)=>(B,S,N*D)
        out = rearrange(_input, 'b s n d -> b s (n d)').contiguous()
    elif input_layout == "SBH":
        # (B,S,N,D)=>(S,B,N*D)
        out = rearrange(_input, 'b s n d -> s b (n d)').contiguous()
    elif input_layout == "BNSD":
        # (B,S,N,D)=>(B,N,S,D)
        out = rearrange(_input, 'b s n d -> b n s d').contiguous()
    elif input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")
    else:
        out = _input
    return out


def convert_to_bsnd(_input, n, input_layout):
    """
    transform qkv from input_layout to bsnd.
    B: batch_size
    S: sequence_length
    N: num_heads
    D: head_dim
    Args:
        _input (torch.Tensor): tensor of shape (B,N,S,D) or (B,S,N,D) or (S,B,H) or (B,S,H)
        n (int): num_heads
        input_layout (str):"BSH" or "SBH" or "BSND" or "BNSD" or "TND"
    Returns:
        tensor of shape (B,S,N,D)
    """
    if input_layout == "BSH":
        # (B,S,N*D)=>(B,S,N,D)
        out = rearrange(_input, 'b s (n d) -> b s n d', n=n)
    elif input_layout == "SBH":
        # (S,B,N*D)=>(B,S,N,D)
        out = rearrange(_input, 's b (n d) -> b s n d', n=n)
    elif input_layout == "BNSD":
        # (B,N,S,D)=>(B,S,N,D)
        out = rearrange(_input, 'b n s d -> b s n d', n=n)
    elif input_layout == "TND":
        raise ValueError(f"input_layout {input_layout} does not supported for now.")
    else:
        out = _input
    if out.dim() != 4:
        raise ValueError(f"convert qkv format failed with input_layout {input_layout}.")
    return out


def generate_atten_mask(*args):
    """
    # 当sparse_mode=2、3、4时小算子到融合算子会走这个优化，反过来看就要拆解回原来的基本实现
    ===> atten_mask = torch.from_numpy(np.triu(np.ones([2048, 2048]), k=1)).to(dtype)
    """

    sparse_mode, atten_mask, b, n1, s1, s2, pre_tocken, next_tocken, dtype = args
    shape = [s1, s2]

    if atten_mask is not None:
        # 当FA的输入已经包含atten_mask时，可以认为已经是转换之后的mask矩阵了，有三种特殊场景，即稀疏矩阵场景，需要进行逆向还原
        if sparse_mode == 2 or sparse_mode == 3 or sparse_mode == 4:
            logger.info(f"s1: {s1}, s2:{s2}, atten_mask.shape:{atten_mask.shape}, atten_mask.dtype:{atten_mask.dtype}")

            if atten_mask.dim() == 2 and atten_mask.shape[0] == 2048 and atten_mask.shape[1] == 2048:
                if atten_mask.equal(torch.from_numpy(np.triu(np.ones([2048, 2048]), k=1)).to(atten_mask.dtype)):
                    if sparse_mode == 2:
                        atten_mask = torch.from_numpy(np.triu(np.ones(shape), k=1))
                    elif sparse_mode == 3:
                        atten_mask = torch.from_numpy(np.triu(np.ones(shape), k=s2 - s1 + 1))
                    elif sparse_mode == 4:
                        atten_mask_u = torch.from_numpy(np.triu(np.ones(shape), k=next_tocken + 1))
                        atten_mask_l = torch.from_numpy(np.tril(np.ones(shape), k=-pre_tocken - 1))
                        atten_mask = atten_mask_u + atten_mask_l
                    logger.debug(f"反向转换atten_mask {atten_mask.shape}")
                    return atten_mask.to(dtype)

        return atten_mask.to(dtype)

    if atten_mask is not None:
        if atten_mask.dim() == 2:
            if atten_mask.shape[0] != s1 or atten_mask.shape[1] != s2:
                raise ValueError(f"Invalid atten_mask shape `SS` {atten_mask.shape}")
            shape = [s1, s2]
        elif atten_mask.dim() == 4:
            if atten_mask.shape[1] == 1:
                shape = [b, 1, s1, s2] if b != 1 else [1, 1, s1, s2]
            else:
                shape = [b, n1, s1, s2] if b != 1 else [1, n1, s1, s2]

    if sparse_mode == 0:
        atten_mask_u = torch.from_numpy(np.triu(np.ones(shape), k=next_tocken + 1))
        atten_mask_l = torch.from_numpy(np.tril(np.ones(shape), k=-pre_tocken - 1))
        atten_mask = atten_mask_u + atten_mask_l
    elif sparse_mode == 1:  # no sparse
        atten_mask = torch.from_numpy(np.zeros(shape))
    elif sparse_mode == 2:
        atten_mask = torch.from_numpy(np.triu(np.ones(shape), k=1))
    elif sparse_mode == 3:
        atten_mask = torch.from_numpy(np.triu(np.ones(shape), k=s2 - s1 + 1))
    elif sparse_mode == 4:
        atten_mask_u = torch.from_numpy(np.triu(np.ones(shape), k=next_tocken + 1))
        atten_mask_l = torch.from_numpy(np.tril(np.ones(shape), k=-pre_tocken - 1))
        atten_mask = atten_mask_u + atten_mask_l
    # 注:不会出现sparse_mode=5的情况，该情况要求必须要传入atten_mask，且atten_mask矩阵数据格式须为BNSS或B1SS，
    # 因此可以认为FA的输入已经是正确的atten_mask了
    return atten_mask.to(dtype)


def generate_kv(key, value, n1, n2):
    # N不等长适配by cdy
    if not (n1 == n2):
        k_new = broadcast_kv(n1, n2, key, key.dtype)
        v_new = broadcast_kv(n1, n2, value, value.dtype)
    else:
        k_new = key
        v_new = value
    return k_new, v_new


def rebuid_softmax_by_qkv(q, k, atten_mask, pse, scale):
    """
    attention = softmax(QK^T/sqrt(d))V
    softmax(x_i) = e^(x_i - x_max) / sum(e^(x_i - x_max))
    """
    logger.info("Using QKV to rebuild original softmax")
    qk = calculate_qk(q, k, atten_mask, pse, scale)
    softmax_res, _, _ = softmax_forward(qk)
    return softmax_res


def rebuild_softmax_by_max_sum(softmax_params):
    """
    attention = softmax(QK^T/sqrt(d))V
    softmax(x_i) = e^(x_i - x_max_i) / x_sum_i)
    """
    q = softmax_params.q
    k = softmax_params.k
    atten_mask = softmax_params.atten_mask
    pse = softmax_params.pse
    scale = softmax_params.scale
    softmax_max = softmax_params.softmax_max
    softmax_sum = softmax_params.softmax_sum
    logger.info("Using softmax_max and softmax_sum to rebuild original softmax")
    qk = calculate_qk(q, k, atten_mask, pse, scale)
    if softmax_max.shape[-1] == 0:
        raise ValueError(f"softmax_max.shape[-1] must be non-zero, softmax_max.shape: {softmax_max.shape}")
    repeat_dim = qk.shape[-1] // softmax_max.shape[-1]
    softmax_res = torch.exp(qk.sub(softmax_max.repeat(1, 1, 1, repeat_dim))).div(
        softmax_sum.repeat(1, 1, 1, repeat_dim))
    return softmax_res


def get_head_num(*args, **kwargs):
    if kwargs.get("head_num", None):
        head_num = kwargs.get("head_num")
    elif len(args) >= 4:
        head_num = args[3]
    else:
        raise ValueError(f"Unsupported npu_fusion_attention args {args}.")
    return head_num


def get_input_layout(*args, **kwargs):
    if kwargs.get("input_layout", None):
        input_layout = kwargs.get("input_layout")
    elif len(args) >= 5:
        input_layout = args[4]
    else:
        raise ValueError(f"Unsupported npu_fusion_attention args {args}.")
    return input_layout


def npu_fusion_attention_forward_patch(*args, **kwargs):

    if len(args) < 2:
        raise RuntimeError("npu_fusion_attention_forward_patch: length of args should greater than or equal to 2.")

    # query, key, value, head_num, input_layout
    head_num = get_head_num(*args, **kwargs)
    input_layout = get_input_layout(*args, **kwargs)

    b, s1, s2, n1, n2, d, h1, h2, dtype = parse_bsnd_args(args[0], args[1], head_num, input_layout)
    if d == 0:
        raise ValueError("npu_fusion_attention_forward_patch: head dimension (d) is zero, division by zero risk.")
    if n1 == n2 and s1 == s2:
        logger.debug(f"running case : BNSD = {b}_{n1}_{s1}_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    else:
        logger.debug(f"running case: BNSD = {b}_{n1}({n2})_{s1}({s2})_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    if not (n1 % n2 == 0 and n1 >= n2):
        raise ValueError(f"N1与N2不匹配,请检查: n1 = {n1}, n2 = {n2}.")

    dims_kwargs = {
        "b": b, "s1": s1, "s2": s2, "n1": n1, "n2": n2,
        "d": d, "h1": h1, "h2": h2, "dtype": dtype
    }

    new_kwargs = {
        "keep_prob": 1,
        "scale": kwargs.get("scale", 1 / (d ** 0.5)),
        "sparse_mode": kwargs.get("sparse_mode", 0),
        "prefix": kwargs.get("prefix"),
        "pre_tockens": kwargs.get("pre_tockens", 2147483647),
        "next_tockens": kwargs.get("next_tockens", 2147483647),
        "pse": kwargs.get("pse"),
        "padding_mask": kwargs.get("padding_mask"),
        "atten_mask": kwargs.get("atten_mask")
    }

    return args, dims_kwargs, new_kwargs


def npu_fusion_attention_backward_patch(*args, **kwargs):
    if len(args) != 6:
        raise ValueError(f"Unsupported npu_fusion_attention_grad args {args}.")

    b, s1, s2, n1, n2, d, h1, h2, dtype = parse_bsnd_args(args[0], args[1], args[4], args[5])
    if d == 0:
        raise ValueError("npu_fusion_attention_backward_patch: head dimension (d) is zero, division by zero risk.")
    if n1 == n2 and s1 == s2:
        logger.info(f"running case : bnsd = {b}_{n1}_{s1}_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    else:
        logger.info(f"running case: bnsd = {b}_{n1}({n2})_{s1}({s2})_{d}, sparse = {kwargs.get('sparse_mode', 0)}")
    if not (n1 % n2 == 0 and n1 >= n2):
        raise ValueError(f"N1与N2不匹配,请检查: n1 = {n1}, n2 = {n2}.")

    dims_kwargs = {
        "b": b, "s1": s1, "s2": s2, "n1": n1, "n2": n2,
        "d": d, "h1": h1, "h2": h2, "dtype": dtype
    }

    new_kwargs = {
        "keep_prob": 1,
        "scale_value": kwargs.get("scale_value", 1 / (d ** 0.5)),
        "sparse_mode": kwargs.get("sparse_mode", 0),
        "prefix": kwargs.get("prefix"),
        "pre_tockens": kwargs.get("pre_tockens", 2147483647),
        "next_tockens": kwargs.get("next_tockens", 2147483647),
        "pse": kwargs.get("pse"),
        "padding_mask": kwargs.get("padding_mask"),
        "softmax_max": kwargs.get("softmax_max"),
        "softmax_sum": kwargs.get("softmax_sum"),
        "softmax_in": kwargs.get("softmax_in"),
        "attention_in": kwargs.get("attention_in"),
        "seed": kwargs.get("seed", 0),
        "offset": kwargs.get("offset", 0),
        "numels": kwargs.get("numels", 0),
        "atten_mask": kwargs.get("atten_mask")
    }

    return args, dims_kwargs, new_kwargs


def npu_fusion_attention(*args, **kwargs):
    new_args, dims_kwargs, new_kwargs = npu_fusion_attention_forward_patch(*args, **kwargs)
    query, key, value = new_args[0], new_args[1], new_args[2]
    input_layout = get_input_layout(*args, **kwargs)
    n1 = dims_kwargs.get("n1")
    n2 = dims_kwargs.get("n2")
    s1 = dims_kwargs.get("s1")
    s2 = dims_kwargs.get("s2")
    b = dims_kwargs.get("b")
    dtype = dims_kwargs.get("dtype")
    atten_mask = new_kwargs.get("atten_mask")
    keep_prob = new_kwargs.get("keep_prob")
    sparse_mode = new_kwargs.get("sparse_mode")
    pre_tockens = new_kwargs.get("pre_tockens")
    next_tockens = new_kwargs.get("next_tockens")
    pse = new_kwargs.get("pse")
    scale = new_kwargs.get("scale")
    args_temp = [sparse_mode, atten_mask, b, n1, s1, s2, pre_tockens, next_tockens, dtype]
    atten_mask = generate_atten_mask(*args_temp)
    query = convert_to_bnsd(query, n1, input_layout)
    key = convert_to_bnsd(key, n2, input_layout)
    value = convert_to_bnsd(value, n2, input_layout)
    k_new, v_new = generate_kv(key, value, n1, n2)
    forward_params = FaForwardParams(query, k_new, v_new, None, atten_mask, pse, scale, keep_prob)
    out_golden, softmax_max, softmax_sum = fusion_attention_forward(forward_params)
    if out_golden.dim() == 5:
        out_golden = out_golden.reshape(out_golden.size(0), out_golden.size(1) * out_golden.size(2), out_golden.size(3),
                                        out_golden.size(4))
    out_golden = convert_from_bnsd(out_golden, input_layout)

    return out_golden.cpu(), softmax_max.repeat(1, 1, 1, 8).cpu(), softmax_sum.repeat(1, 1, 1, 8).cpu()


def npu_fusion_attention_grad(*args, **kwargs):
    # dx, q, k, v, softmax_res, drop_mask, pse, scale, keep_prob
    new_args, dims_kwargs, new_kwargs = npu_fusion_attention_backward_patch(*args, **kwargs)
    query, key, value, dx, input_layout = new_args[0], new_args[1], new_args[2], new_args[3], new_args[5]
    n1 = dims_kwargs.get("n1")
    n2 = dims_kwargs.get("n2")
    s1 = dims_kwargs.get("s1")
    s2 = dims_kwargs.get("s2")
    b = dims_kwargs.get("b")
    d = dims_kwargs.get("d")
    dtype = dims_kwargs.get("dtype")
    atten_mask = new_kwargs.get("atten_mask")
    keep_prob = new_kwargs.get("keep_prob")
    sparse_mode = new_kwargs.get("sparse_mode")
    pre_tockens = new_kwargs.get("pre_tockens")
    next_tockens = new_kwargs.get("next_tockens")
    pse = new_kwargs.get("pse")
    softmax_max = new_kwargs.get("softmax_max")
    softmax_sum = new_kwargs.get("softmax_sum")
    scale_value = new_kwargs.get("scale_value")

    args_temp = [sparse_mode, atten_mask, b, n1, s1, s2, pre_tockens, next_tockens, dtype]
    atten_mask = generate_atten_mask(*args_temp)
    query = convert_to_bnsd(query, n1, input_layout)
    dx = convert_to_bnsd(dx, n1, input_layout)
    key = convert_to_bnsd(key, n2, input_layout)
    value = convert_to_bnsd(value, n2, input_layout)
    k_new, v_new = generate_kv(key, value, n1, n2)

    if SOFTMAX_BUILD_MODE == "QKV":
        softmax_res = rebuid_softmax_by_qkv(query, k_new, atten_mask, pse, scale_value)
    else:
        softmax_params = RebuildSoftmaxParams(query, k_new, atten_mask, pse, scale_value, softmax_max, softmax_sum)
        softmax_res = rebuild_softmax_by_max_sum(softmax_params)
    backward_params = FaBackwardParams(dx, query, k_new, v_new, softmax_res, None, pse, scale_value, keep_prob)
    dq, dk, dv = fusion_attention_backward(backward_params)

    # N不等长适配by cdy
    if not (n1 == n2):
        if n2 == 0:
            raise ValueError("dims_kwargs.n2 must be non-zero.")
        g = int(n1 / n2)
        dk = torch.sum(dk.reshape(b, n2, g, s2, d), dim=2, keepdim=True).reshape(b, n2, s2, d)
        dv = torch.sum(dv.reshape(b, n2, g, s2, d), dim=2, keepdim=True).reshape(b, n2, s2, d)

    if dq.dim() == 5:
        dq = dq.reshape(dq.size(0), dq.size(1) * dq.size(2), dq.size(3), dq.size(4))
    if dk.dim() == 5:
        dk = dk.reshape(dk.size(0), dk.size(1) * dk.size(2), dk.size(3), dk.size(4))
    if dv.dim() == 5:
        dv = dv.reshape(dv.size(0), dv.size(1) * dv.size(2), dv.size(3), dv.size(4))

    dq = convert_from_bnsd(dq, input_layout)
    dk = convert_from_bnsd(dk, input_layout)
    dv = convert_from_bnsd(dv, input_layout)

    return dq.cpu(), dk.cpu(), dv.cpu()


def is_attention_off_due_to_mask(atten_mask_dtype):
    return not atten_mask_dtype


def is_attention_off_in_sparse_mode_4(sparse_mode, next_tockens, pre_tockens, s1):
    return sparse_mode == 4 and (next_tockens != 0 or pre_tockens < s1)


def is_attention_off_in_sparse_mode_0(sparse_mode, pre_tockens, next_tockens, s1, s2):
    return sparse_mode == 0 and pre_tockens >= s1 and next_tockens >= s2


def gpu_fusion_attention(*args, **kwargs):
    deterministic = False
    new_args, dims_kwargs, new_kwargs = npu_fusion_attention_forward_patch(*args, **kwargs)
    query, key, value = new_args[0], new_args[1], new_args[2]
    keep_prob = new_kwargs.get("keep_prob", 1.0)
    scale = new_kwargs.get("scale")
    n1 = dims_kwargs.get("n1")
    n2 = dims_kwargs.get("n2")
    s1 = dims_kwargs.get("s1")
    s2 = dims_kwargs.get("s2")
    b = dims_kwargs.get("b")
    pse = new_kwargs.get("pse")
    sparse_mode = new_kwargs.get("sparse_mode")
    pre_tockens = new_kwargs.get("pre_tockens")
    next_tockens = new_kwargs.get("next_tockens")
    attn_mask = new_kwargs.get("atten_mask")
    atten_mask_dtype = attn_mask.dtype if new_kwargs.get("atten_mask") is not None else None
    pre_tockens = min(CompareConst.MAX_TOKENS, pre_tockens)
    next_tockens = min(CompareConst.MAX_TOKENS, next_tockens)
    atten_off = (is_attention_off_due_to_mask(atten_mask_dtype) or
                 is_attention_off_in_sparse_mode_4(sparse_mode, next_tockens, pre_tockens, s1) or
                 is_attention_off_in_sparse_mode_0(sparse_mode, pre_tockens, next_tockens, s1, s2))
    causal_switch = not atten_off
    if sparse_mode == CompareConst.SPECIAL_SPARSE_MOED:
        window_left = pre_tockens
        window_right = next_tockens
    else:
        pre_tockens = next_tockens = CompareConst.MAX_TOKENS
        window_left = pre_tockens - s1 + s2
        window_right = next_tockens + s1 - s2

    if pse is not None:
        alibi_slopes = torch.rand(b, n1, dtype=torch.float32) * 0.3
    else:
        alibi_slopes = None

    input_layout = get_input_layout(*args, **kwargs)
    query = convert_to_bsnd(query, n1, input_layout)
    key = convert_to_bsnd(key, n2, input_layout)
    value = convert_to_bsnd(value, n2, input_layout)
    out = flash_attn_func(
        query, key, value, dropout_p=(1 - keep_prob), softmax_scale=scale, causal=causal_switch,
        window_size=(window_left, window_right), alibi_slopes=alibi_slopes, deterministic=deterministic
    )
    out = convert_from_bsnd(out, input_layout)
    return out, Const.NONE, Const.NONE



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Any

from msprobe.infer.utils.check import NumberChecker, ObjectChecker, StringChecker, DictChecker, PathChecker, ListChecker
from msprobe.infer.utils.check.checker import Checker
from msprobe.infer.utils.constants import INPUT_FILE_MAX_SIZE


class NumConverter:
    def __init__(self, convert_type=float):
        self.convert_type = convert_type

    def __call__(self, value) -> Any:
        return self.convert(value)

    def convert(self, value: str):
        try:
            if self.convert_type == int:
                return int(value), True, ""
            elif self.convert_type == float:
                return float(value), True, ""
            else:
                return value, True, ""
        except ValueError as er:
            return value, False, str(er)


class Rule:
    @staticmethod
    def none() -> Checker:
        return Checker().is_none()

    @staticmethod
    def num() -> NumberChecker:
        return NumberChecker().is_number()

    @staticmethod
    def str() -> StringChecker:
        return StringChecker().is_str()

    @staticmethod
    def dict() -> DictChecker:
        return DictChecker().is_dict()

    @staticmethod
    def obj(obj_type) -> ObjectChecker:
        return ObjectChecker().is_type(obj_type)

    @staticmethod
    def path() -> PathChecker:
        return PathChecker()

    @staticmethod
    def list() -> ListChecker:
        return ListChecker().is_list()

    @staticmethod
    def config_file() -> PathChecker:
        return (
            PathChecker()
            .exists()
            .is_file()
            .is_readable()
            .is_not_writable_to_others()
            .is_safe_parent_dir()
            .max_size(10 * 1000 * 1000)
            .as_default()
        )

    @staticmethod
    def input_file() -> PathChecker:
        return (
            PathChecker()
            .exists()
            .forbidden_softlink()
            .is_file()
            .is_readable()
            .is_owner()
            .is_not_writable_to_others()
            .is_safe_parent_dir()
            .max_size(INPUT_FILE_MAX_SIZE)
            .as_default()
        )

    @staticmethod
    def input_dir() -> PathChecker:
        return PathChecker().exists().is_dir().is_readable().is_uid_matched().is_not_writable_to_others().as_default()

    @staticmethod
    def output_dir() -> PathChecker:
        return (
            Rule.path()
            .any(Rule.anti(PathChecker().exists()), PathChecker().is_dir().is_writeable().is_not_writable_to_others())
            .as_default()
        )

    @staticmethod
    def any(*rules: Checker) -> Checker:
        return Checker().any(*rules)

    @staticmethod
    def anti(rule: Checker) -> Checker:
        return Checker().anti(rule)

    @staticmethod
    def to_int() -> NumberChecker:
        return NumberChecker(converter=NumConverter(int))

    @staticmethod
    def to_float() -> NumberChecker:
        return NumberChecker(converter=NumConverter(float))



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch
import numpy as np


def softmax_func(x, axis=None):
    x = x.float()
    x_max = x.max(dim=axis, keepdims=True).values
    x_sub = x - x_max
    y = torch.exp(x_sub)
    x_sum = y.sum(dim=axis, keepdims=True)
    ans = 0 if (x_sum == 0).any() else y / x_sum
    return ans


def npu_moe_gating_top_k_softmax(x, finished_optional, k):
    input_dtype = x.dtype
    if x.dim() < 1:
        raise ValueError("Input x must have at least 1 dimensions.")
    num_expert = x.shape[-1]
    softmax = softmax_func(x, -1)
    softmax = softmax.to(input_dtype)
    expert_idx = torch.argsort(-softmax, dim=-1, stable=True)
    expert_idx = expert_idx[:, :k]
    y = torch.gather(softmax, index=expert_idx, dim=-1)
    if finished_optional is not None:
        if finished_optional.dim() < 1:
            raise ValueError("Finished_optional must have at least 1 dimensions.")
        finished_optional = finished_optional.view(finished_optional.shape[0], 1)
        finished_optional = finished_optional.expand(-1, k)
        expert_idx = torch.where(finished_optional, num_expert, expert_idx)
    if y.dim() < 2:
        raise ValueError("Variable y must have at least 2 dimensions.")
    row_idx = torch.arange(y.shape[0] * y.shape[1]).reshape(y.shape[1], y.shape[0]).t()

    return y, expert_idx, row_idx



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
convert format from NDC1HWC0 to ND.
"""
import numpy as np


def convert(shape_from: list, shape_to: list, array: any) -> any:
    """
    Convert the data format from NDC1HWC0 to ND. Current only support 5 dimensions ND
    :param shape_from: the shape before convert
    :param shape_to: the shape after convert
    :param array: the one-dimensional array
    :return:the data array of ND shape
    """
    axis_n, axis_d, axis_c1, axis_h, axis_w, axis_c0 = shape_from

    tmp_array = array.reshape(shape_from)
    tmp_array = np.transpose(tmp_array, axes=(0, 2, 5, 1, 3, 4))
    # first convert to NCDHW
    tmp_array = tmp_array.reshape((axis_n, axis_c1 * axis_c0, axis_d, axis_h, axis_w))

    if len(shape_to) == 5:
        c_pad = None if axis_c1 * axis_c0 == shape_to[1] else shape_to[1] - axis_c1 * axis_c0
        return tmp_array[:, :c_pad, ...]
    else:
        return tmp_array



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch


def npu_scaled_masked_softmax(x, mask, scale, fixed_triu_mask):
    if fixed_triu_mask:
        mask = (torch.triu(torch.ones(mask.shape), k=1)).bool().to(mask.device)
    dtype = x.dtype
    x = (x * scale).masked_fill(mask, value=-10000)
    x = x - torch.max(x, dim=-1, keepdims=True)[0]
    x = torch.exp(x.float())
    y = torch.div(x, torch.sum(x, dim=-1, keepdims=True))
    return y.to(dtype)


def npu_scaled_masked_softmax_backward(y_grad, y, mask, scale, fixed_triu_mask):
    if fixed_triu_mask:
        mask = (torch.triu(torch.ones(mask.shape), k=1)).bool().to(mask.device)
    dtype = y_grad.dtype
    y_grad = y_grad.float()
    y = y.float()
    x_grad = y_grad * y
    x_grad = y_grad - torch.sum(x_grad, dim=-1, keepdims=True)
    x_grad = x_grad * y
    x_grad = x_grad * scale
    x_grad = x_grad.masked_fill(mask, value=0)
    return x_grad.to(dtype).cpu()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves xxxx function.
"""
import os
import re
import time

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.file_utils import FileUtils
from msprobe.msaccucmp.cmp_utils import path_check as path_utils
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.compare_vector import VectorComparison
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_rule_parser import FusionRuleParser


class BatchCompare:
    """
    The class for batch compare
    """
    DUMP_FILE_PATH_FORMAT = "dump_path/time/device_id/model_name/model_id/dump_step/dump_file"

    def __init__(self: any) -> None:
        self.model_name_to_json_map = {}
        self.json_path_to_dump_path_map = {}

    @staticmethod
    def _check_path_valid(path: str) -> None:
        ret = path_utils.check_path_valid(path, True, False, path_utils.PathType.Directory)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)

    def check_fusion_rule_json_dir(self: any, fusion_rule_json_dir: str) -> bool:
        """
        Check fusion rule is dir or not
        :param fusion_rule_json_dir: the path for fusion rule json
        :return:bool
        """
        fusion_rule_path = os.path.realpath(fusion_rule_json_dir)
        if os.path.isfile(fusion_rule_path):
            return False
        self._check_path_valid(fusion_rule_json_dir)
        return True

    def check_argument_valid(self: any, arguments: any) -> None:
        """
        Check argument valid
        """
        if arguments.op_name:
            log.print_error_log(
                "the {} single operator comparison is not supported for batch network comparison.".format(
                    arguments.op_name))
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

        npu_dump_path = os.path.realpath(arguments.my_dump_path)
        self._check_path_valid(npu_dump_path)
        # check whether the path ends with a timestamp.
        path = os.path.split(npu_dump_path)
        # path[1]:last level directory
        if not path[1].isdigit():
            log.print_error_log("The {} path must end with a timestamp.".format(npu_dump_path))
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)

    def compare(self: any, arguments: any) -> int:
        """
        Compare the entire network in batches
        :param arguments: the command parameters
        :return: the compare finish status code
        """
        self.check_argument_valid(arguments)
        self._make_model_name_to_json_map(os.path.realpath(arguments.fusion_rule_file))
        self._make_json_path_to_dump_path_map(os.path.realpath(arguments.my_dump_path))
        return self._execute_batch_compare(arguments)

    def _make_model_name_to_json_map(self: any, json_dir_path: str) -> None:
        for json_file_name in os.listdir(json_dir_path):
            json_file_path = os.path.join(json_dir_path, json_file_name)
            if json_file_path.endswith(".json"):
                self._parse_json_file(json_file_path)

        if not self.model_name_to_json_map:
            log.print_error_log('There is no fusion rule json file in "%r".' % json_dir_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)

    def _parse_json_file(self: any, json_file_path: str) -> None:
        fusion_rule_parse = FusionRuleParser(json_file_path)
        json_object = FileUtils.load_json_file(json_file_path)
        fusion_rule_parse.check_array_object_valid(json_object, ConstManager.GRAPH_OBJECT)
        for graph in json_object[ConstManager.GRAPH_OBJECT]:
            fusion_rule_parse.check_string_object_valid(graph, ConstManager.NAME_OBJECT)
            model_name = graph[ConstManager.NAME_OBJECT]
            self.model_name_to_json_map[model_name] = json_file_path

    def _make_map_for_inconsistent_timestamp(self: any, graph_name: str, dump_file_path_map: dict,
                                             npu_dump_dir: str) -> None:
        # 1.get the number list from graph name.
        # graph_name:'ge_default_20210420113943_73',graph_name_number_list:['20210420113943', '73'].
        graph_name_number_list = re.findall(r"\d+", graph_name)
        for _, model_name in enumerate(dump_file_path_map):
            # 1.get the number list from model name.for example:
            # model_name:'ge_default_20210420113944_73',model_name_number_list:['20210420113944', '73'].
            model_name_number_list = re.findall(r"\d+", model_name)
            # check graph_name and model_name inconsistent timestamps
            if len(graph_name_number_list) != 2 or len(graph_name_number_list) != len(model_name_number_list):
                log.print_npu_path_valid_message(npu_dump_dir, self.DUMP_FILE_PATH_FORMAT)
                raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
            if graph_name_number_list[1] == model_name_number_list[1]:
                dump_file_path = dump_file_path_map.get(model_name)
                log.print_warn_log(
                    "The {0} and in {1} dump data are not generated at the same time. "
                    "The result may be incorrect.".format(graph_name, ",".join(dump_file_path)))
                self.json_path_to_dump_path_map[self.model_name_to_json_map.get(graph_name)] = dump_file_path

    def _make_json_path_to_dump_path_map(self: any, npu_dump_dir: str) -> None:
        dump_file_path_map = self._get_npu_dump_file_map(npu_dump_dir)
        graph_name_list = self.model_name_to_json_map.keys()
        for graph_name in graph_name_list:
            dump_file_path = dump_file_path_map.get(graph_name)
            if dump_file_path:
                self.json_path_to_dump_path_map[self.model_name_to_json_map.get(graph_name)] = dump_file_path
            else:
                self._make_map_for_inconsistent_timestamp(graph_name, dump_file_path_map, npu_dump_dir)

        if not self.json_path_to_dump_path_map:
            log.print_error_log("The {} not match fusion rule file".format(npu_dump_dir))
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)

    def _execute_batch_compare(self: any, arguments: any) -> int:
        if os.path.islink(os.path.abspath(arguments.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % arguments.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        output_path = os.path.realpath(arguments.output_path)
        my_dump_path = os.path.realpath(arguments.my_dump_path)
        ret = CompareError.MSACCUCMP_NONE_ERROR
        for key, value in self.json_path_to_dump_path_map.items():
            for dump_path in value:
                path_data = dump_path.replace(my_dump_path, "").split("/")
                # remove model_id from path_data,path_data:['0', 'ge_default_20210420113943_71', '10', '1']
                model_id_index = len(path_data) - 2
                path_data.pop(model_id_index)
                if "" in path_data:
                    path_data.remove("")
                result_csv = 'result_%s.csv' % time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
                # generate result file,for example:'0_ge_default_20210420113943_71_1_result_20210715022552.csv'
                file_name = "".join(["_".join(path_data), "_", result_csv])
                arguments.my_dump_path = dump_path
                arguments.fusion_rule_file = key
                compare = VectorComparison(arguments)
                compare.set_output_path(os.path.join(output_path, file_name))
                ret = compare.compare()
        return ret

    def _get_npu_dump_file_map_by_model_id(self: any, model_name_dir: str, dump_file_path_map: dict) -> None:
        for model_id in os.listdir(model_name_dir):
            model_id_dir = os.path.join(model_name_dir, model_id)
            self._check_path_valid(model_id_dir)
            model_id_list = os.listdir(model_id_dir)
            model_name = os.path.basename(model_name_dir)
            if len(model_id_list) != 0:
                step = max(model_id_list) if len(model_id_list) > 1 else model_id_list[0]
                dump_step_path = os.path.join(model_id_dir, step)
                if dump_file_path_map.get(model_name):
                    dump_file_path_map.get(model_name).append(dump_step_path)
                else:
                    dump_file_path_map[model_name] = [dump_step_path]

    def _get_npu_dump_file_map(self: any, npu_dump_dir: str) -> dict:
        dump_file_path_map = {}
        for device_id in os.listdir(npu_dump_dir):
            device_id_dir = os.path.join(npu_dump_dir, device_id)
            self._check_path_valid(device_id_dir)
            for model_name in os.listdir(device_id_dir):
                model_name_dir = os.path.join(device_id_dir, model_name)
                self._check_path_valid(model_name_dir)
                self._get_npu_dump_file_map_by_model_id(model_name_dir, dump_file_path_map)
        if not dump_file_path_map:
            log.print_npu_path_valid_message(npu_dump_dir, self.DUMP_FILE_PATH_FORMAT)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        return dump_file_path_map



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from tensorboard.util import tb_logging
from ..utils.global_state import GraphState
from ..utils.constant import NPU_PREFIX, BENCH_PREFIX, NPU, SINGLE, UNEXPAND_NODE, MODULE, DataType

logger = tb_logging.get_logger()

INNER_WIDTH = 50  # 内部节点宽度
INNER_HIGHT = 15  # 内部节点高度
HORIZONTAL_SPACING = 5  # 横向排列间距
VERTICAL_SPACING = 10  # 纵向排列间距
MAX_PER_ROW = 5  # 横向每行最大数
DB_TYPE = DataType.DB.value
JSON_TYPE = DataType.JSON.value


class Hierarchy:
    
    def __init__(self, graph_type, repo, micro_step, rank, step):
        # DB：查询根节点信息
        node_info = repo.query_root_nodes(graph_type, rank, step)
        if not node_info:
            logger.error("No root node info found in database.")
            return
        root_node_name = node_info.get('node_name')
        name_prefix = NPU_PREFIX if graph_type == NPU else BENCH_PREFIX
        name_prefix = '' if graph_type == SINGLE else name_prefix
        self.repo = repo
        self.name_prefix = name_prefix
        self.root_name = root_node_name
        self.graph_type = graph_type
        self.rank = rank
        self.step = step
        self.micro_step_id = micro_step
        self.current_hierarchy = {
            root_node_name: self.get_basic_rende_info(root_node_name, node_info)
        }
        # 默认展开根节点
        self.update_graph_data(self.root_name)
        self.update_graph_shape()
        self.update_graph_position()

    @classmethod
    def measure_text_width(cls, text):
        return len(text) * 6  # 假设每个字符宽度为6

    @classmethod
    def extract_label_name(cls, node_name, node_type):
        if not node_name:
            return ''
        splited_subnode_name = node_name.split('.')
        splited_label = []
        # 在展开层级时，将父级层级名称相关去除，仅保留子节点本身名称信息
        # 如Module.layer1.1.relu.ReLU.forward.0中的父级名称Module.layer1.1去除，仅保留子级的relu.ReLU.forward.1
        # 如Module.layer4.0.BasicBlock.forward.0中的父级名称Module.1去除，仅保留子级的layer4.0.BasicBlock.forward.0
        if int(node_type) == MODULE:
            if len(splited_subnode_name) < 4:
                return node_name
            elif not splited_subnode_name[-4].isdigit():
                splited_label = splited_subnode_name[-4:]
            elif len(splited_subnode_name) > 4:
                splited_label = splited_subnode_name[-5:]
            else:
                return node_name
        # 在展开层级时，将父级层级名称相关去除，仅保留API子节点本身名称信息，
        # 如 Module.layer1.1.ApiList.1 中的父级名称Module.layer1.1去除，仅保留子级的ApiList.1
        # 如 Module.layer1.1.ApiList.0.1 中的父级名称Module.layer1.1去除，仅保留子级的ApiList.0.1
        else:
            if len(splited_subnode_name) < 2:
                return node_name
            elif not splited_subnode_name[-2].isdigit():
                splited_label = splited_subnode_name[-2:]
            elif len(splited_subnode_name) > 2:
                splited_label = splited_subnode_name[-3:]
            else:
                return node_name
        return ('.').join(splited_label)

    def update_graph_data(self, node_name):
        target_node = self.current_hierarchy.get(node_name, {})
        if node_name == self.root_name or target_node:
            self.process_click_expand(node_name)
        else:  # 如果图中不存在该节点，说明是选中节点，需要递归展开该节点的所有父节点
            self.process_select_expand(node_name)

    def process_click_expand(self, node_name):
        target_node = self.current_hierarchy.get(node_name, {})
        target_node_children = target_node.get("children", [])
        if not target_node or not target_node_children:
            return
        if not target_node.get('expand', False):
            # 1.将target_node的expand置为true
            # 2.将node_name的子节点信息初始化，并添加到current_hierarchy中
            # DB: 查询所有以当前为父节点的子节点
            sub_nodes = self.repo.query_sub_nodes(node_name, self.graph_type, self.rank, self.step)
            for subnode_name, node_info in sub_nodes.items():
                if self.current_hierarchy.get(subnode_name):
                    continue
                render_info = self.get_basic_rende_info(subnode_name, node_info)
                self.current_hierarchy[subnode_name] = render_info
            target_node['expand'] = True
        else:
            target_node['expand'] = False if node_name != self.root_name else True  # 根节点默认展开

    def process_select_expand(self, node_name):
        # DB：逻辑
        # 1.查询当前节点及其的所有父节点，一直到没有父节点位置{}
        
        # DB：查询当前节点的父节点信息
        up_nodes = self.repo.query_up_nodes(node_name, self.graph_type, self.rank, self.step)
        parent_node_name = up_nodes.get(node_name, {}).get('upnode')
        parent_node = self.current_hierarchy.get(parent_node_name, '')
        # 递归展开父节点
        while not parent_node or not parent_node.get('expand', False):
            if not parent_node:  # 如果父节点不存在图中，则初始化父节点
              
                render_info = self.get_basic_rende_info(parent_node_name, up_nodes.get(parent_node_name))
                self.current_hierarchy[parent_node_name] = render_info
            try:
                self.process_click_expand(parent_node_name)  # 展开父节点
            except Exception as e:
                logger.error(f"Failed to expand parent node {parent_node_name}: {e}")
                break
            if parent_node_name == self.root_name:
                break
            parent_node_name = up_nodes.get(parent_node_name, {}).get('upnode')
            parent_node = self.current_hierarchy.get(parent_node_name, '')

    def update_graph_shape(self):
        self.resize_hierarchy(self.root_name)

    def update_graph_position(self):
        self.layout_hierarchy(self.root_name)

    def resize_hierarchy(self, current_name):
        node = self.current_hierarchy.get(current_name, {})
        if not node:
            return
        if node.get('nodeType') == UNEXPAND_NODE:
            # 不可展开叶子节点固定尺寸
            node['width'] = INNER_WIDTH
            node['height'] = INNER_HIGHT
            return
        if not node.get('expand', False):
            # 未展开的父节点按文字宽度
            node['width'] = self.measure_text_width(node.get('label', '')) + HORIZONTAL_SPACING * 2  # 文字宽度 + 边距
            node['height'] = INNER_HIGHT
            return
        for child_name in node.get('children', []):
            self.resize_hierarchy(child_name)
        max_child_width = 0
        total_height = INNER_HIGHT
        for group_type, children in self.group_children(node.get('children')):
            if group_type == UNEXPAND_NODE:
                # 横向布局：分批处理每行最多MAX_PER_ROW个
                rows = [children[i:i + MAX_PER_ROW] for i in range(0, len(children), MAX_PER_ROW)]
                if (len(rows) > 1):
                    max_child_width = max(max_child_width,
                                          INNER_WIDTH * MAX_PER_ROW + (HORIZONTAL_SPACING * (MAX_PER_ROW - 1)))
                    total_height += ((INNER_HIGHT + VERTICAL_SPACING) * len(rows))
                else:
                    child_total_width = sum(self.current_hierarchy.get(child, {}).get('width', 0) for child in children)
                    spacing = HORIZONTAL_SPACING * (len(children) - 1)
                    max_child_width = max(max_child_width, child_total_width + spacing)
                    total_height += (INNER_HIGHT + VERTICAL_SPACING)

            else:
                # 纵向布局，计算子节点最大宽度
                # 更新 max_child_width：取所有子节点中最宽的一个
                child_max_width = max(self.current_hierarchy.get(child, {}).get('width', 0) for child in children)
                max_child_width = max(max_child_width, child_max_width)
                # 累加 total_height：所有子节点高度 + 垂直间距（每个子节点之间都有间距，所以是 len(children) 个间距）
                children_heights = sum(self.current_hierarchy.get(child, {}).get('height', 0) for child in children)
                vertical_spacing = VERTICAL_SPACING * len(children)
                total_height += (children_heights + vertical_spacing)

        # 最终尺寸计算
        node['width'] = max(
            max_child_width + HORIZONTAL_SPACING * 2,  # 子节点最大宽度 + 边距
            self.measure_text_width(node.get('label', '')) + HORIZONTAL_SPACING * 2  # 保证文字可见
        )
        node['height'] = total_height + VERTICAL_SPACING  # 总高度 + 边距

    def layout_hierarchy(self, current_name):
        node = self.current_hierarchy.get(current_name, {})
        if not node or not node.get('children'):
            return
        parent_x = node.get('x')
        parent_width = node.get('width')
        current_y = node.get('y') + INNER_HIGHT + VERTICAL_SPACING  # 初始Y坐标

        # 按类型分组处理子节点
        for group_type, children in self.group_children(node.get('children')):
            if group_type == UNEXPAND_NODE:
                # 横向布局：分批处理每行最多MAX_PER_ROW个
                rows = [children[i:i + MAX_PER_ROW] for i in range(0, len(children), MAX_PER_ROW)]

                for row_children in rows:
                    # 计算该行总宽度
                    row_width = sum(
                        self.current_hierarchy.get(child_name, {}).get('width') for child_name in row_children)
                    row_width += HORIZONTAL_SPACING * (len(row_children) - 1)

                    # 居中对齐计算
                    start_x = parent_x + (parent_width - row_width) // 2
                    max_height = 0
                    current_x = start_x

                    # 布局该行所有子节点
                    for child_name in row_children:
                        child = self.current_hierarchy.get(child_name, {})
                        child['x'] = current_x
                        child['y'] = current_y
                        current_x += child.get('width') + HORIZONTAL_SPACING
                        max_height = max(max_height, child.get('height'))

                    # 更新Y坐标
                    current_y += max_height + VERTICAL_SPACING
            else:
                # 纵向布局：每个子节点单独占一行且居中
                for child_name in children:
                    child = self.current_hierarchy.get(child_name, {})
                    child['x'] = parent_x + (parent_width - child.get('width')) // 2
                    child['y'] = current_y
                    current_y += child.get('height') + VERTICAL_SPACING
                    if (child.get('expand', False)):
                        self.layout_hierarchy(child_name)  # 递归处理子节点

    def group_children(self, children):
        """将子节点按类型分组，连续同类型节点为一组"""
        groups = []
        current_group = []
        current_type = None

        for name in children:
            child = self.current_hierarchy.get(name, {})
            if not child:
                continue
            # 类型变化时创建新组
            if current_type is not None and child.get('nodeType') != current_type:
                groups.append((current_type, current_group))
                current_group = []
            current_type = child.get('nodeType')
            current_group.append(name)
        # 添加最后一组
        if current_group:
            groups.append((current_type, current_group))
        return groups

    def get_basic_rende_info(self, node_name, node_info):
        if not node_info:
            return {}
        label = node_name
        children = []
        precision_index = None,
        if node_name == self.root_name:  # 根节点，根据micro_step获取子节点
            # DB：查询以根节点为父节点的所有子节点
            sub_nodes = self.repo.query_sub_nodes(node_name, self.graph_type, self.rank, self.step)
            for subnode_name, child_node in sub_nodes.items():
                child_micro_step_id = child_node.get('micro_step_id', -1)  # 如果子节点不包含micro_step_id，则默认为-1，直接添加
                is_append_all_node = int(self.micro_step_id) == -1 or child_micro_step_id == -1
                is_append_split_node = int(self.micro_step_id) != -1 and int(child_micro_step_id) == int(
                    self.micro_step_id)
                if is_append_all_node or is_append_split_node:
                    children.append(subnode_name)
        else:
            children = node_info.get('subnodes', [])
        # precisionIndex 先从缓存中获取(更新误差可能会更新缓存)，如果不存在，则从ode_info中获取
        update_precision_cache = GraphState.get_global_value('update_precision_cache', {})
        if update_precision_cache.get(node_name) is not None:
            precision_index = update_precision_cache.get(node_name, {}).get('precision_index')
        else:
            precision_index = node_info.get('data', {}).get('precision_index', "NaN")
        if node_info.get('upnode', '') != self.root_name:  # 首层节点不处理显示内容
            label = self.extract_label_name(node_name, node_info.get('node_type'))
        render_info = {
            'x': 0,
            'y': 0,
            'width': self.measure_text_width(node_name) + HORIZONTAL_SPACING * 2,
            'height': INNER_HIGHT,
            'expand': False,
            'isRoot': node_name == self.root_name,
            'parentNode': node_info.get('upnode', ''),
            'label': label,
            'name': self.name_prefix + node_name,
            'children': children,
            'nodeType': node_info.get('node_type') if node_info.get("subnodes") else UNEXPAND_NODE,
            'matchedNodeLink': node_info.get('matched_node_link', []),
            'precisionIndex': precision_index,  # 精度
            'overflowLevel': node_info.get('data', {}).get('overflow_level', "NaN"),  # 溢出
            'matchedDistributed': node_info.get('matched_distributed', {}),
        }
        return render_info

    # 获取连通图
    def get_connected_graph(self, name, result, new_hierarchy):
        node = self.current_hierarchy.get(name, {})
        if (not node): 
            return
        filtered_value = {k: v for k, v in node.items() if k != "children"}
        result[name] = filtered_value
        new_hierarchy[name] = node
        # 递归处理子节点
        if (node.get('expand')):
            for child_name in node.get('children', []):
                self.get_connected_graph(child_name, result, new_hierarchy)

    def update_hierarchy_data(self):
        if self.repo.repo_type == DB_TYPE:
            return self.current_hierarchy
        # 处理JSON
        for node_name, node_info in self.current_hierarchy.items():
            graph_node_info = self.repo.query_node_info(node_name, self.graph_type)
            node_info['matchedNodeLink'] = graph_node_info.get('matched_node_link', [])
            node_info['precisionIndex'] = graph_node_info.get('data', {}).get('precision_index', "NaN")
        return self.current_hierarchy

    def update_current_hierarchy_data(self, data):
        for node_info in data:
            node_name = node_info.get('node_name')
            if node_name in self.current_hierarchy:
                current_node_info = self.current_hierarchy[node_name]
                current_node_info['matchedNodeLink'] = node_info.get('matched_node_link')
                current_node_info['precisionIndex'] = node_info.get('precision_index')
        return True
            
    def get_hierarchy(self):
        result = {}
        new_hierarchy = {}
        # 遍历当前层级结构
        self.get_connected_graph(self.root_name, result, new_hierarchy)
        self.current_hierarchy = new_hierarchy  # 折叠之后的需要重新展开
        return result



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.file_utils import create_directory, save_json
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.dump.debugger.debugger_config import DebuggerConfig


class KernelGraphDump:
    def __init__(self, config: DebuggerConfig):
        self.dump_json = dict()
        self.dump_json["common_dump_settings"] = dict()
        self.dump_json["common_dump_settings"]["dump_mode"] = 0
        self.dump_json["common_dump_settings"]["path"] = ""
        self.dump_json["common_dump_settings"]["net_name"] = "Net"
        self.dump_json["common_dump_settings"]["iteration"] = "all"
        self.dump_json["common_dump_settings"]["saved_data"] = "statistic"
        self.dump_json["common_dump_settings"]["input_output"] = 0
        self.dump_json["common_dump_settings"]["kernels"] = []
        self.dump_json["common_dump_settings"]["support_device"] = [0, 1, 2, 3, 4, 5, 6, 7]
        self.dump_json["common_dump_settings"]["op_debug_mode"] = 0
        self.dump_json["common_dump_settings"]["file_format"] = "npy"

        if len(config.list) > 0:
            self.dump_json["common_dump_settings"]["dump_mode"] = 1
            self.dump_json["common_dump_settings"]["kernels"] = config.list
        self.dump_json["common_dump_settings"]["path"] = config.dump_path
        if len(config.step) > 0:
            step_str = ""
            for s in config.step:
                step_str += (str(s) + '|')
            self.dump_json["common_dump_settings"]["iteration"] = step_str[:-1]
        if len(config.rank) > 0:
            self.dump_json["common_dump_settings"]["support_device"] = config.rank
        if config.task == "tensor":
            self.dump_json["common_dump_settings"]["saved_data"] = "tensor"
            self.dump_json["common_dump_settings"]["file_format"] = config.file_format
        if len(config.data_mode) == 1:
            if config.data_mode[0] == "input":
                self.dump_json["common_dump_settings"]["input_output"] = 1
            if config.data_mode[0] == "output":
                self.dump_json["common_dump_settings"]["input_output"] = 2

    def handle(self):
        try:
            from msprobe.lib import _msprobe_c
            return
        except ImportError:
            # 如果没有_msprobe_ce_c走MindSpore老流程
            logger.info("Module _msprobe_c has not been installed, use interface in mindspore instead.")

        if os.getenv("GRAPH_OP_RUN") == "1":
            raise Exception("Must run in graph mode, not kbk mode")
        json_path = self.dump_json["common_dump_settings"]["path"]
        create_directory(json_path)
        json_path = os.path.join(json_path, "kernel_graph_dump.json")
        save_json(json_path, self.dump_json, indent=4)
        logger.info(json_path + " has been created.")
        os.environ["MINDSPORE_DUMP_CONFIG"] = json_path
        if self.dump_json["common_dump_settings"]["dump_mode"] == 0:
            if self.dump_json["common_dump_settings"]["iteration"] != "all" or \
               len(self.dump_json["common_dump_settings"]["kernels"]) == 0:
                os.environ["MS_ACL_DUMP_CFG_PATH"] = json_path
        else:
            if "MS_ACL_DUMP_CFG_PATH" in os.environ:
                del os.environ["MS_ACL_DUMP_CFG_PATH"]
